"""Interface adapter for the Generalized Linear Model Regressor."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["ShreeshaM07", "julian-fong"]

import numpy as np
import pandas as pd

from skpro.regression.base import BaseProbaRegressor


class GLMRegressor(BaseProbaRegressor):
    """
    Fits a generalized linear model with a gaussian link.

    Direct interface to ``statsmodels.genmod.generalized_linear_model.GLM``
    from the ``statsmodels`` package.

    For a direct link to statmodels' Generalized Linear Models module see:
    https://www.statsmodels.org/stable/glm.html#module-reference

    Parameters
    ----------
    family : string, default : "Normal"
        The family parameter denotes the type of distribution
        that will be used.
        Available family/distributions are
        1."Normal"
        2."Poisson"
        3."Gamma"
    link : string, default : None
        This parameter is used to represent the link function to be
        used with the distribution.
        If default is None it will internally replace with default of the
        respective family. The default is the first string
        against each family below.
        Available safe options for the respective family are:
        ``Normal`` : "Identity", "Log", "InversePower";
        ``Poisson`` : "Log", "Identity", "Sqrt";
        ``Gamma`` : "InversePower", "Log", "Identity";
    offset_var : string or int, default = None
        Pass the column name as a string or column number as an int in X.
        If string, then the exog or ``X`` passed while ``fit``-ting
        must have an additional column with column name passed through
        ``offset_var`` with any values against each row. When ``predict``ing
        have an additional column with name same as string passed through ``offset_var``
        in X with all the ``offset_var`` values for predicting
        stored in the column for each row.
        If ``int`` it corresponding column number will be considered.
    exposure_var : string or int, default = None
        Pass the column name as a string or column number as an int in X.
        If string, then the exog or ``X`` passed while ``fit``-ting
        must have an additional column with column name passed through
        ``exposure_var`` with any values against each row. When ``predict``ing
        have additional column with name same as string passed through ``exposure_var``
        in X with all the ``exposure_var`` values for predicting
        stored in the column for each row.
        If ``int`` it corresponding column number will be considered.
    missing : str
        Available options are 'none', 'drop' and 'raise'. If 'none', no nan
        checking is done. If 'drop', any observations with nans are dropped.
        If 'raise', an error is raised. Default = 'none'
    start_params : array_like (optional)
        Initial guess of the solution for the loglikelihood maximization.
        The default is family-specific and is given by the
        family.starting_mu(endog). If start_params is given then the initial
        mean will be calculated as np.dot(exog, start_params).
        This parameter is used inside the GLM fit() function.

    maxiter : int, optional, default=100
        Number of iterations. This parameter is used inside the GLM fit() function.

    method : str, optional, default='IRLS'
        Default is 'IRLS' for iteratively re-weighted least squares.
        This parameter is used inside the GLM fit() function.

    tol : float, optional, default=1e-8
        Convergence tolerance. Default is 1e-8. This parameter is
        used inside the GLM fit() function.

    scale : str/float, optional, default=None
        scale can be 'X2', 'dev', or a float. The default value is None,
        which uses X2 for gamma, gaussian and inverse gaussian. X2 is
        Pearson's chi-square divided by df_resid. The default is 1 for
        the Bionmial and Poisson families. dev is the deviance divided
        by df_resid. This parameter is used inside the GLM fit() function.

    cov_type : str, optional, default='nonrobust'
        The type of parameter estimate covariance matrix to compute.
        This parameter is used inside the GLM fit() function.

    cov_kwds : dict-like, optional, default=None
        Extra arguments for calculating the covariance of the
        parameter estimates. This parameter is used inside the GLM fit() function.

    use_t : bool, optional, default=False
        if True, the Student t-distribution if used for inference.
        This parameter is used inside the GLM fit() function.

    full_output : bool, optional, default=True
        Set to True to have all available output in the Results object’s
        mle_retvals attribute. The output is dependent on the solver. See
        LikelihoodModelResults notes section for more information. Not used
        if methhod is IRLS. This parameter is used inside the GLM fit() function.

    disp : bool, optional, default=False
        Set to True to print convergence messages. Not used if method
        is IRLS. This parameter is used inside the GLM fit() function.

    max_start_irls : int, optional, default=3
        The number of IRLS iterations used to obtain starting values for
        gradient optimization. Only relevenat if method is set to something
        other than "IRLS". This parameter is used inside the GLM fit() function.

    add_constant : bool, optional, default=False
        statsmodels does not include an intercept by default. Specify this as
        True if you would like to add an intercept (floats of 1s) to the
        dataset X. Default = False. Note that when the input is a pandas
        Series or DataFrame, the added column's name is 'const'.

    Attributes
    ----------
    df_model_ : float
        Model degrees of freedom is equal to p - 1, where p is the number of
        regressors. Note that the intercept is not reported as a degree of freedom.

    df_resid_ : float
        Residual degrees of freedom is equal to the number of observation n
        minus the number of regressors p.

    iteration_ : int
        The number of iterations that fit has run. Initialized at 0.
        Only available after fit is called

    mu_ : ndarray
        The mean response of the transformed variable. mu is the value of the
        inverse of the link function at lin_pred, where lin_pred is the linear
        predicted value of the WLS fit of the transformed variable. mu is only
        available after fit is called. See statsmodels.families.family.fitted
        of the distribution family for more information.

    n_trials_ : ndarray
        Note that n_trials is a reference to the data so that if data is
        already an array and it is changed, then n_trials changes as well.
        n_trials is the number of binomial trials and only available with that
        distribution. See statsmodels.families.Binomial for more information.

    scale_ : float
        The estimate of the scale / dispersion of the model fit.
        Only available after fit is called. See GLM.fit and GLM.estimate_scale
        for more information.

    scaletype_ : str
        The scaling used for fitting the model. This is only available
        after fit is called. The default is None. See GLM.fit for
        more information.

    weights_ : ndarray
        The value of the weights after the last iteration of fit.
        Only available after fit is called. See statsmodels.families.family
        for the specific distribution weighting functions.

    glm_fit_ : GLM
        fitted generalized linear model

    fit_history_ : dict
        Contains information about the iterations.
        Its keys are iterations, deviance and params. Only available after
        fit is called.

    model_ : class instance
        Pointer to GLM model instance that called fit.

    nobs_ : float
        The number of observations n. Only available after fit is called.

    normalized_cov_params_ : ndarray
        For Gaussian link: This is the p x p normalized covariance of the
        design / exogenous data. This is approximately equal to (X.T X)^(-1)

    params_ : ndarray
        The coefficients of the fitted model. Note that interpretation of the
        coefficients often depends on the distribution family and the data.

    pvalues_ : ndarray
        The two-tailed p-values for the parameters.

    scale_ : float
        The estimate of the scale / dispersion for the model fit.
        See GLM.fit and GLM.estimate_scale for more information.

    stand_errors_ : ndarray
        The standard errors of the fitted GLM.
    """

    _tags = {
        "authors": ["ShreeshaM07", "julian-fong"],
        "maintainers": ["ShreeshaM07", "julian-fong"],
        "python_version": None,
        "python_dependencies": "statsmodels",
        "capability:multioutput": False,
        "capability:missing": False,
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
    }

    def _str_to_sm_family(self, family, link):
        """Convert the string to a statsmodel object.

        If the link function is also explcitly mentioned then include then
        that must be passed to the family/distribution object.
        """
        from warnings import warn

        from statsmodels.genmod.families.family import Gamma, Gaussian, Poisson
        from statsmodels.genmod.families.links import Identity, InversePower, Log, Sqrt

        sm_fmly = {
            "Normal": Gaussian,
            "Poisson": Poisson,
            "Gamma": Gamma,
        }

        links = {
            "Log": Log,
            "Identity": Identity,
            "InversePower": InversePower,
            "Sqrt": Sqrt,
        }

        if link in links:
            link_function = links[link]()
            try:
                return sm_fmly[family](link_function)
            except Exception:
                msg = "Invalid link for family, default link will be used"
                warn(msg)

        return sm_fmly[family]()

    def __init__(
        self,
        family="Normal",
        link=None,
        offset_var=None,
        exposure_var=None,
        missing="none",
        start_params=None,
        maxiter=100,
        method="IRLS",
        tol=1e-8,
        scale=None,
        cov_type="nonrobust",
        cov_kwds=None,
        use_t=None,
        full_output=True,
        disp=False,
        max_start_irls=3,
        add_constant=False,
    ):
        super().__init__()

        self.family = family
        self.link = link
        self.offset_var = offset_var
        self.exposure_var = exposure_var
        self.missing = missing
        self.start_params = start_params
        self.maxiter = maxiter
        self.method = method
        self.tol = tol
        self.scale = scale
        self.cov_type = cov_type
        self.cov_kwds = cov_kwds
        self.use_t = use_t
        self.full_output = full_output
        self.disp = disp
        self.max_start_irls = max_start_irls
        self.add_constant = add_constant

        self._family = self.family
        self._link = self.link
        self._offset_var = self.offset_var
        self._exposure_var = self.exposure_var
        self._missing = self.missing
        self._start_params = self.start_params
        self._maxiter = self.maxiter
        self._method = self.method
        self._tol = self.tol
        self._scale = self.scale
        self._cov_type = self.cov_type
        self._cov_kwds = self.cov_kwds
        self._use_t = self.use_t
        self._full_output = self.full_output
        self._disp = self.disp
        self._max_start_irls = self.max_start_irls
        self._add_constant = self.add_constant

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            A (n,k) array where n is the number of observations and k is the number
            of regressors. An intercept is not included by default and should be
            added by the user (models specified using a formula include an
            intercept by default). Equivalent to statsmodel's (exog).

        y : pandas DataFrame
            1d array of the endogenous response variable. This array can be 1d or
            2d. Binomial family models accept a 2d array  with two columns.
            If supplied each observation is expected to  be [success, failure].
            Equivalent to statsmodel's (endog).

        Returns
        -------
        self : reference to self
        """
        from statsmodels.genmod.generalized_linear_model import GLM

        # remove the offset and exposure columns which
        # was inserted to maintain the shape
        offset_var = self._offset_var
        exposure_var = self._exposure_var

        X_ = self._prep_x(X, offset_var, exposure_var, False)

        y_col = y.columns

        family = self._family
        link = self._link
        sm_family = self._str_to_sm_family(family=family, link=link)

        glm_estimator = GLM(
            endog=y,
            exog=X_,
            family=sm_family,
            missing=self._missing,
        )

        self._estimator = glm_estimator

        glm_fit_params = {
            "start_params": self._start_params,
            "maxiter": self._maxiter,
            "method": self._method,
            "tol": self._tol,
            "scale": self._scale,
            "cov_type": self._cov_type,
            "cov_kwds": self._cov_kwds,
            "use_t": self._use_t,
            "full_output": self._full_output,
            "disp": self._disp,
            "max_start_irls": self._max_start_irls,
        }

        fitted_glm_model = glm_estimator.fit(**glm_fit_params)

        PARAMS_TO_FORWARD = {
            "df_model_": glm_estimator.df_model,
            "df_resid_": glm_estimator.df_resid,
            "mu_": glm_estimator.mu,
            "n_trials_": glm_estimator.n_trials,
            "weights_": glm_estimator.weights,
            "scaletype_": glm_estimator.scaletype,
        }

        for k, v in PARAMS_TO_FORWARD.items():
            setattr(self, k, v)

        # forward some parameters to self
        FITTED_PARAMS_TO_FORWARD = {
            "glm_fit_": fitted_glm_model,
            "y_col": y_col,
            "fit_history_": fitted_glm_model.fit_history,
            "iteration_": fitted_glm_model.fit_history["iteration"],
            "model_": fitted_glm_model.model,
            "nobs_": fitted_glm_model.nobs,
            "normalized_cov_params_": fitted_glm_model.normalized_cov_params,
            "params_": fitted_glm_model.params,
            "pvalues_": fitted_glm_model.pvalues,
            "scale_": fitted_glm_model.scale,
            "stand_errors_": fitted_glm_model.bse,
        }

        for k, v in FITTED_PARAMS_TO_FORWARD.items():
            setattr(self, k, v)

        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted"

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, with same columns as y in fit
        """
        offset_var = self._offset_var
        exposure_var = self._exposure_var
        offset_arr = None
        exposure_arr = None

        X_, offset_arr, exposure_arr = self._prep_x(X, offset_var, exposure_var, True)

        index = X_.index
        y_column = self.y_col
        y_pred_series = self.glm_fit_.predict(
            X_, offset=offset_arr, exposure=exposure_arr
        )
        y_pred = pd.DataFrame(y_pred_series, index=index, columns=y_column)

        return y_pred

    def _params_sm_to_skpro(self, y_predictions_df, index, columns, family):
        """Convert the statsmodels output to equivalent skpro distribution."""
        from skpro.distributions.gamma import Gamma
        from skpro.distributions.normal import Normal
        from skpro.distributions.poisson import Poisson

        skpro_distr = {
            "Normal": Normal,
            "Poisson": Poisson,
            "Gamma": Gamma,
        }

        params = {}
        skp_dist = Normal

        if family in skpro_distr:
            skp_dist = skpro_distr[family]

        if skp_dist == Normal:
            y_mu = y_predictions_df["mean"].rename("mu").to_frame()
            y_sigma = y_predictions_df["mean_se"].rename("sigma").to_frame()
            params["mu"] = y_mu
            params["sigma"] = y_sigma
        elif skp_dist == Poisson:
            y_mu = y_predictions_df["mean"].rename("mu").to_frame()
            params["mu"] = y_mu
        elif skp_dist == Gamma:
            y_mean = y_predictions_df["mean"]
            y_sd = y_predictions_df["mean_se"]
            y_alpha = (y_mean / y_sd) ** 2
            y_beta = (y_mean / (y_sd**2)).rename("beta").to_frame()
            y_alpha = y_alpha.rename("alpha").to_frame()
            params["alpha"] = y_alpha
            params["beta"] = y_beta

        params["index"] = index
        params["columns"] = columns

        y_pred = skp_dist(**params)
        return y_pred

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y_pred : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        # remove the offset and exposure columns
        # which was inserted to maintain the shape
        offset_var = self._offset_var
        exposure_var = self._exposure_var

        X_ = self._prep_x(X, offset_var, exposure_var, False)

        # instead of using the conventional predict() method, we use statsmodels
        # get_prediction method, which returns a pandas df that contains
        # the prediction and prediction variance i.e mu and sigma
        y_column = self.y_col
        y_predictions_df = self.glm_fit_.get_prediction(X_).summary_frame()

        # convert the returned values to skpro equivalent distribution
        family = self._family
        index = X_.index
        columns = y_column

        y_pred = self._params_sm_to_skpro(y_predictions_df, index, columns, family)
        return y_pred

    def _prep_x(self, X, offset_var, exposure_var, rtn_off_exp_arr):
        """
        Return a copy of X with an added constant of self.add_constant = True.

        If rtn_off_exp_arr is True it will also return offset and exposure
        arrays along with updated X.

        Parameters
        ----------
        X : pandas DataFrame
            Dataset that the user is trying to do inference on

        Returns
        -------
        X_ : pandas DataFrame
            A copy of the input X with an added column 'const' with is an
            array of len(X) of 1s
        offset_arr : numpy.array
            The copy of column which is meant for offsetting present in X.
        exposure_arr : numpy.array
            The copy of column which is meant for exposure present in X.
        """
        from statsmodels.tools import add_constant

        offset_arr = None
        exposure_arr = None
        if offset_var is not None:
            if isinstance(offset_var, str):
                offset_var = pd.Index([offset_var])
                offset_arr = np.array(X[offset_var]).flatten()
            elif isinstance(offset_var, int):
                offset_arr = np.array(X.iloc[:, offset_var]).flatten()
                offset_var = pd.Index([X.iloc[:, offset_var].name])
        if exposure_var is not None:
            if isinstance(exposure_var, str):
                exposure_var = pd.Index([exposure_var])
                exposure_arr = np.array(X[exposure_var]).flatten()
            elif isinstance(exposure_var, int):
                exposure_arr = np.array(X.iloc[:, exposure_var]).flatten()
                exposure_var = pd.Index([X.iloc[:, exposure_var].name])
        # drop the offset and exposure columns from X
        columns_to_drop = []
        if offset_var is not None:
            columns_to_drop.append(offset_var[0])
        if exposure_var is not None:
            columns_to_drop.append(exposure_var[0])
        if columns_to_drop:
            X = X.drop(columns_to_drop, axis=1)

        if self._add_constant:
            X_ = add_constant(X)
            if rtn_off_exp_arr:
                return X_, offset_arr, exposure_arr
            return X_
        else:
            if rtn_off_exp_arr:
                return X, offset_arr, exposure_arr
            return X

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}
        params2 = {"add_constant": True}
        params3 = {
            "family": "Poisson",
            "add_constant": True,
        }
        params4 = {"family": "Gamma"}
        params5 = {
            "family": "Normal",
            "link": "InversePower",
        }
        params6 = {
            "family": "Poisson",
            "link": "Log",
            "add_constant": True,
        }

        return [params1, params2, params3, params4, params5, params6]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Configure skpro Sphinx documentation."""

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import datetime
import inspect
import os
import sys

import skpro

# -- Path setup --------------------------------------------------------------

# When we build the docs on readthedocs, we build the package and want to
# use the built files in order for sphinx to be able to properly read the
# Cython files. Hence, we do not add the source code path to the system path.
env_rtd = os.environ.get("READTHEDOCS")
# Check if on Read the docs
if not env_rtd == "True":
    sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------

current_year = datetime.datetime.now().year
project = "skpro"
project_copyright = f"2017 - {current_year} (BSD-3-Clause License)"
author = "skpro Developers"


# The full version, including alpha/beta/rc tags
release = skpro.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "numpydoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",  # link to GitHub source code via linkcode_resolve()
    "nbsphinx",  # integrates example notebooks
    "sphinx_gallery.load_style",
    "myst_parser",
    "sphinx_panels",
    "sphinx_issues",
    "sphinx_design",
]

myst_enable_extensions = ["colon_fence"]

# -- Internationalization ------------------------------------------------
# specifying the natural language populates some key tags
language = "en"

# Use bootstrap CSS from theme.
panels_add_bootstrap_css = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# The main toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# add_module_names = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# Members and inherited-members default to showing methods and attributes from
# a class or those inherited.
# Member-order orders the documentation in the order of how the members are
# defined in the source code.
autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "member-order": "bysource",
}

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# Link to GitHub repo for github_issues extension
issues_github_path = "sktime/skpro"


def linkcode_resolve(domain, info):
    """Return URL to source code corresponding.

    Parameters
    ----------
    domain : str
    info : dict

    Returns
    -------
    url : str
    """

    def find_source():
        # try to find the file and line number, based on code from numpy:
        # https://github.com/numpy/numpy/blob/main/doc/source/conf.py#L286
        obj = sys.modules[info["module"]]
        for part in info["fullname"].split("."):
            obj = getattr(obj, part)

        fn = inspect.getsourcefile(obj)
        fn = os.path.relpath(fn, start=os.path.dirname(skpro.__file__))
        source, lineno = inspect.getsourcelines(obj)
        return fn, lineno, lineno + len(source) - 1

    if domain != "py" or not info["module"]:
        return None
    try:
        filename = "skpro/%s#L%d-L%d" % find_source()
    except Exception:
        filename = info["module"].replace(".", "/") + ".py"
    if version_match == "latest":
        version = "main"
    else:
        version = version_match
    return f"https://github.com/sktime/skpro/blob/{version}/{filename}"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "pydata_sphinx_theme"

# Define the json_url for our version switcher.
json_url = "https://skpro.readthedocs.io/en/latest/_static/switcher.json"

# This uses code from the py-data-sphinx theme's own conf.py
# Define the version we use for matching in the version switcher.
version_match = os.environ.get("READTHEDOCS_VERSION")

# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# If it is an integer, we're in a PR build and the version isn't correct.
if not version_match or version_match.isdigit():
    # For local development, infer the version to match from the package.
    if "dev" in release or "rc" in release:
        version_match = "latest"
        # We want to keep the relative reference if we are in dev mode
        # but we want the whole url if we are effectively in a released version
        json_url = "_static/switcher.json"
    else:
        version_match = "v" + release

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/sktime/skpro",
            "icon": "fab fa-github",
        },
        {
            "name": "Discord",
            "url": "https://discord.com/invite/54ACzaFsn7",
            "icon": "fab fa-discord",
        },
        {
            "name": "LinkedIn",
            "url": "https://www.linkedin.com/company/scikit-time/",
            "icon": "fab fa-linkedin",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/skpro",
            "icon": "fa-solid fa-box",
        },
    ],
    "icon_links_label": "Quick Links",
    "show_nav_level": 1,
    "show_prev_next": False,
    "use_edit_page_button": False,
    "navbar_start": ["navbar-logo", "version-switcher"],
    "navbar_center": ["navbar-nav"],
    "switcher": {
        "json_url": json_url,
        "version_match": version_match,
    },
    "header_links_before_dropdown": 6,
}
html_logo = "images/skpro-banner.png"
html_context = {
    "github_user": "sktime",
    "github_repo": "skpro",
    "github_version": "main",
    "doc_path": "docs/source/",
    "default_mode": "light",
}
html_sidebars = {"**": ["sidebar-nav-bs.html", "sidebar-ethical-ads.html"]}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ---------------------------------------------
# Output file base name for HTML help builder.
htmlhelp_basename = "skprodoc"


# -- Options for LaTeX output ------------------------------------------------
# latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
# 'papersize': 'letterpaper',
# The font size ('10pt', '11pt' or '12pt').
# 'pointsize': '10pt',
# Additional stuff for the LaTeX preamble.
# 'preamble': '',
# Latex figure (float) alignment
# 'figure_align': 'htbp',
# }

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "skpro.tex",
        "skpro Documentation",
        "skpro developers",
        "manual",
    ),
]

# -- Extension configuration -------------------------------------------------

# -- Options for numpydoc extension ------------------------------------------
# see http://stackoverflow.com/q/12206334/562769
numpydoc_show_class_members = True
# this is needed for some reason...
# see https://github.com/numpy/numpydoc/issues/69
numpydoc_class_members_toctree = False

numpydoc_validation_checks = {"all", "GL01", "SA01", "EX01"}

# -- Options for sphinx-copybutton extension----------------------------------
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for nbsphinx extension ------------------------------------------
nbsphinx_execute = "never"  # always  # whether to run notebooks
nbsphinx_allow_errors = False  # False
nbsphinx_timeout = 600  # seconds, set to -1 to disable timeout

# add Binder launch button at the top
current_file = "{{ env.doc2path( env.docname, base=None) }}"

# make sure Binder points to latest stable release, not main
binder_base = "https://mybinder.org/v2/gh//skpro/"
binder_url = binder_base + f"{version_match}?filepath={current_file}"
nbsphinx_prolog = f"""
.. |binder| image:: https://mybinder.org/badge_logo.svg
.. _Binder: {binder_url}

|Binder|_
"""

# add link to original notebook at the bottom
notebook_url = f"https://github.com/sktime/skpro/tree/{version_match}/{current_file}"
nbsphinx_epilog = f"""
----

Generated using nbsphinx_. The Jupyter notebook can be found here_.

.. _here: {notebook_url}
.. _nbsphinx: https://nbsphinx.readthedocs.io/
"""

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": (f"https://docs.python.org/{sys.version_info.major}", None),
    "scikit-learn": ("https://scikit-learn.org/stable/", None),
    "sktime": ("https://www.sktime.net/en/stable/", None),
}



"""Reduction to classification using sklearn classifiers fit on binned data."""

__author__ = ["ShreeshaM07"]

import numpy as np
import pandas as pd

from skpro.distributions import Histogram
from skpro.regression.base import BaseProbaRegressor
from skpro.utils.sklearn import prep_skl_df


class HistBinnedProbaRegressor(BaseProbaRegressor):
    """A binned probabilistic regressor fitting a histogram distribution.

    It is a probabilistic regressor that fits a Histogram Distribution
    by presenting binned outcomes to a probabilistic sklearn classifier.
    It can be used for predicting the class that a set of X belongs to and
    predict_proba can be used to represent the predicted probabilites of
    each class for respective values in X in the form of a Histogram
    Distribution.

    The ``bins`` will be used to bin the ``y`` values in fit into the respective
    bins. It then uses these bins as the classes for the classifier and predicts
    the probabilites for each class.

    Note: Ensure the ``y`` values while calling ``fit`` are within the the ``bins``
    range. If it is not then it will be internally replaced to move to the
    closest bin.

    Parameters
    ----------
    clf : instance of a sklearn classifier
        Classifier to wrap, must have ``predict`` and ``predict_proba``.
    bins : int or 1D array of float, default: 10
        1. If ``int`` then it will be considered as the number of bins.
        2. Else if it is an array then it will be used as the bin boundaries.
        If the requirement is ``n`` bins then the ``len(bins)`` must be ``n+1``.

    Attributes
    ----------
    classes_ : np.array
        Contains the names of the classes that it was fit on.
    class_bin_map_ : dict
        The key contains the class name assigned to the bin.
        It maps the key (which indicates the ``i``th bin) to the respective
        bin's boundaries np.array([bins[i],bins[i+1]]).
    classes_proba_ : pd.DataFrame
        Contains the class probabilites.
    """

    _tags = {
        "authors": ["ShreeshaM07"],
        "maintainers": ["ShreeshaM07"],
        "capability:multioutput": False,
        "capability:missing": True,
    }

    def __init__(self, clf, bins=10):
        self.clf = clf
        self.bins = bins

        super().__init__()

    def _bins_int_arr(self, bins, y):
        y = np.array(y).flatten()
        start = min(y) * 0.999
        stop = max(y) * 1.001
        bins = np.linspace(start=start, stop=stop, num=bins + 1)
        return bins

    def _y_bins_compatiblity(self, y, bins, _y_cols):
        y = np.array(y).flatten()
        upper_y = bins[-1] - 1e-9
        lower_y = bins[0] + 1e-9
        y = np.where(y <= bins[0], lower_y, y)
        y = np.where(y >= bins[-1], upper_y, y)
        y = pd.DataFrame(y, columns=_y_cols)
        return y

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        from warnings import warn

        from numpy.lib.stride_tricks import sliding_window_view
        from sklearn import clone

        self.clf_ = clone(self.clf)
        bins = self.bins
        self._y_cols = y.columns

        # create bins array in case of bins being an `int`
        if isinstance(bins, int) or isinstance(bins, np.integer):
            bins = self._bins_int_arr(bins, y)

        # check if y values are within bins range
        # if not move it to the closest bin.
        y = self._y_bins_compatiblity(y, bins, self._y_cols)

        # in case of int it will be internally replaced in fit
        self._bins = bins

        # Generate class names based on bins
        class_bins = [f"class{i}" for i in range(len(bins) - 1)]
        self._class_bins = class_bins

        if len(bins) != len(class_bins) + 1:
            warn(
                f"len of `bins` is {len(bins)} != len of classes {len(class_bins)}+1."
                " Ensure the bins has all the bin boundaries resulting in"
                " number of bins + 1 elements."
            )

        bins_hist = sliding_window_view(bins, window_shape=2)
        # maps the bin boundaries [bin start,bin end] to the classes
        class_bin_map_ = {}
        for i in range(len(bins_hist)):
            class_bin_map_[class_bins[i]] = bins_hist[i]
        self.class_bin_map_ = class_bin_map_

        # bins the y values into classes.
        class_series = pd.cut(y.iloc[:, 0], bins=bins, labels=class_bins, right=True)
        y_binned = pd.DataFrame(class_series, columns=self._y_cols)

        X = prep_skl_df(X)
        y_binned = prep_skl_df(y_binned)

        if isinstance(y_binned, pd.DataFrame) and len(y_binned.columns) == 1:
            y_binned = y_binned.iloc[:, 0]
        elif len(y_binned.shape) > 1 and y_binned.shape[1] == 1:
            y_binned = y_binned[:, 0]

        self.clf_.fit(X, y_binned)
        self.classes_ = self.clf_.classes_

        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        X = prep_skl_df(X)
        y_pred = self.clf_.predict(X)
        y_pred_df = pd.DataFrame(y_pred, index=X.index, columns=self._y_cols)
        return y_pred_df

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        from warnings import warn

        X = prep_skl_df(X)
        bins = self._bins
        classes_ = self.classes_
        class_bins = self._class_bins

        if len(bins) != len(class_bins) + 1:
            warn(
                f"len of `bins` is {len(bins)} != len of classes {len(class_bins)}+1."
                " Ensure the bins has all the bin boundaries resulting in"
                " number of bins + 1 elements."
            )

        y_pred_proba = self.clf_.predict_proba(X)
        # map classes probabilities/bin_mass to class names
        classes_proba_ = pd.DataFrame(y_pred_proba, columns=classes_)

        # Identify missing classes
        missing_classes = set(class_bins) - set(classes_)
        if missing_classes:
            # Add missing classes with 0 values
            for missing_class in missing_classes:
                classes_proba_[missing_class] = 0
            # Sort columns based on the numerical part of the class names
            # in order to match with the bins while calling Histogram distribution
            classes_proba_ = classes_proba_.reindex(
                sorted(classes_proba_.columns, key=lambda x: int(x[5:])), axis=1
            )

        self.classes_proba_ = classes_proba_
        y_pred_proba = np.array(classes_proba_)

        if len(X) == 1:
            bin_mass = y_pred_proba[0]
            pred_proba = Histogram(bins=bins, bin_mass=bin_mass)
            return pred_proba

        # converting it to a 2D shape
        bin_mass = np.array([y_pred_proba])
        # Reshape and swap axes to get the desired structure
        bin_mass = bin_mass.swapaxes(0, 1).reshape(-1, 1, bin_mass.shape[-1])

        bins = np.array([[bins]] * len(X))

        pred_proba = Histogram(
            bins=bins, bin_mass=bin_mass, index=X.index, columns=self._y_cols
        )
        return pred_proba

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.naive_bayes import GaussianNB
        from sklearn.semi_supervised import LabelSpreading
        from sklearn.tree import DecisionTreeClassifier

        param1 = {"clf": DecisionTreeClassifier(), "bins": 4}
        param2 = {"clf": GaussianNB()}
        params3 = {"clf": LabelSpreading(), "bins": [20, 80, 160, 250, 300, 380, 420]}

        return [param1, param2, params3]



"""Utility functions for plotting."""
import numpy as np
import pandas as pd
from skbase.utils.dependencies import _check_soft_dependencies

__authors__ = ["fkiraly", "frthjf"]


def plot_crossplot_interval(y_true, y_pred, coverage=None, ax=None):
    """Probabilistic cross-plot for regression, truth vs prediction interval.

    Plots:

    * x-axis: ground truth value
    * y-axis: median predictive value, with error bars being
      the prediction interval at symmetric coverage ``coverage``

    Parameters
    ----------
    y_true : array-like, [n_samples, n_targets]
        Ground truth values
    y_pred : skpro distribution, or predict_interval return, [n_samples, n_targets]
        symmetric prediction intervals are obtained
        via the coverage parameterfrom y_pred
        Predicted values
    coverage : float, optional, default=0.9
        Coverage of the prediction interval
        Used only if y_pred a distribution
    ax : matplotlib axes, optional
        Axes to plot on, if None, a new figure is created and returned

    Returns
    -------
    ax : matplotlib axes
        Axes containing the plot
        If ax was None, a new figure is created and returned
        If ax was not None, the same ax is returned with plot added

    Examples
    --------
    >>> from skpro.utils.plotting import plot_crossplot_interval  # doctest: +SKIP
    >>> from skpro.regression.residual import ResidualDouble  # doctest: +SKIP
    >>> from sklearn.ensemble import RandomForestRegressor  # doctest: +SKIP
    >>> from sklearn.linear_model import LinearRegression  # doctest: +SKIP
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> reg_mean = LinearRegression()  # doctest: +SKIP
    >>> reg_resid = RandomForestRegressor()  # doctest: +SKIP
    >>> reg_proba = ResidualDouble(reg_mean, reg_resid)  # doctest: +SKIP
    >>>
    >>> reg_proba.fit(X, y)  # doctest: +SKIP
    ResidualDouble(...)
    >>> y_pred = reg_proba.predict_proba(X)  # doctest: +SKIP
    >>> plot_crossplot_interval(y, y_pred)  # doctest: +SKIP
    """
    _check_soft_dependencies("matplotlib")

    from matplotlib import pyplot

    if hasattr(y_pred, "quantile") and not isinstance(y_pred, pd.DataFrame):
        if coverage is None:
            coverage = 0.9
        quantile_pts = [0.5 - coverage / 2, 0.5, 0.5 + coverage / 2]
        y_quantiles = y_pred.quantile(quantile_pts)
        y_mid = y_quantiles.iloc[:, 1]
        y_quantiles = y_quantiles.iloc[:, [0, 2]]
    else:
        y_quantiles = y_pred
        y_mid = y_quantiles.mean(axis=1)

    y_mid_two = pd.DataFrame([y_mid, y_mid]).values
    y_quantiles_np = y_quantiles.values.T
    y_bars = np.abs(y_mid_two - y_quantiles_np)

    if ax is None:
        _, ax = pyplot.subplots()

    ax.plot(y_true, y_true, "g.", label="Optimum")
    ax.errorbar(
        y_true.values,
        y_mid,
        yerr=y_bars,
        label="Predictions",
        fmt="b.",
        ecolor="r",
        linewidth=0.5,
    )
    ax.set_ylabel(r"Prediction interval $\widehat{y}_i$")
    ax.set_xlabel(r"Correct label $y_i$")
    ax.legend(loc="best")

    return ax


def plot_crossplot_std(y_true, y_pred, ax=None):
    r"""Probabilistic cross-plot for regression, error vs predictive standard deviation.

    Plots:

    * x-axis: absolute error samples :math:`|y_i - \widehat{y}_i.\mu|`
    * y-axis: predictive standard deviation :math:`\widehat{y}_i.\sigma`,
      of the prediction :math:`\widehat{y}_i` corresponding to :math:`y_i`

    Parameters
    ----------
    y_true : array-like, [n_samples, n_targets]
        Ground truth values
    y_pred : skpro distribution, or predict_var return, [n_samples, n_targets]
        Predicted values
    ax : matplotlib axes, optional
        Axes to plot on, if None, a new figure is created and returned

    Returns
    -------
    ax : matplotlib axes
        Axes containing the plot
        If ax was None, a new figure is created and returned
        If ax was not None, the same ax is returned with plot added

    Examples
    --------
    >>> from skpro.utils.plotting import plot_crossplot_std  # doctest: +SKIP
    >>> from skpro.regression.residual import ResidualDouble  # doctest: +SKIP
    >>> from sklearn.ensemble import RandomForestRegressor  # doctest: +SKIP
    >>> from sklearn.linear_model import LinearRegression  # doctest: +SKIP
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> reg_mean = LinearRegression()  # doctest: +SKIP
    >>> reg_resid = RandomForestRegressor()  # doctest: +SKIP
    >>> reg_proba = ResidualDouble(reg_mean, reg_resid)  # doctest: +SKIP
    >>>
    >>> reg_proba.fit(X, y)  # doctest: +SKIP
    ResidualDouble(...)
    >>> y_pred = reg_proba.predict_proba(X)  # doctest: +SKIP
    >>> plot_crossplot_std(y, y_pred)  # doctest: +SKIP
    """
    _check_soft_dependencies("matplotlib")

    from matplotlib import pyplot

    if hasattr(y_pred, "_tags") and not isinstance(y_pred, pd.DataFrame):
        y_var = y_pred.var()
    else:
        y_var = y_pred

    y_std = np.sqrt(y_var)

    if ax is None:
        _, ax = pyplot.subplots()

    ax.plot(
        np.abs(y_pred.mean().values.flatten() - y_true.values.flatten()),
        y_std.values.flatten(),
        "b.",
    )
    ax.set_ylabel(r"Predictive standard deviation of $\widehat{y}_i$")
    ax.set_xlabel(r"Absolute errors $|y_i - \widehat{y}_i|$")
    # ax.legend(loc="best")

    return ax


def plot_crossplot_loss(y_true, y_pred, metric, ax=None):
    r"""Cross-loss-plot for probabilistic regression.

    Plots:

    * x-axis: ground truth values :math:`y_i`
    * y-axis: loss of the prediction :math:`\widehat{y}_i`
      corresponding to :math:`y_i`,
      as calculated by ``metric.evaluate_by_index``

    Parameters
    ----------
    y_true : array-like, [n_samples, n_targets]
        Ground truth values
    y_pred : skpro distribution, or predict_var return, [n_samples, n_targets]
        Predicted values
    metric : skpro metric
        Metric to calculate the loss
    ax : matplotlib axes, optional
        Axes to plot on, if None, a new figure is created and returned

    Returns
    -------
    ax : matplotlib axes
        Axes containing the plot
        If ax was None, a new figure is created and returned
        If ax was not None, the same ax is returned with plot added

    Examples
    --------
    >>> from skpro.utils.plotting import plot_crossplot_loss  # doctest: +SKIP
    >>> from skpro.metrics import CRPS  # doctest: +SKIP
    >>> from skpro.regression.residual import ResidualDouble  # doctest: +SKIP
    >>> from sklearn.ensemble import RandomForestRegressor  # doctest: +SKIP
    >>> from sklearn.linear_model import LinearRegression  # doctest: +SKIP
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> reg_mean = LinearRegression()  # doctest: +SKIP
    >>> reg_resid = RandomForestRegressor()  # doctest: +SKIP
    >>> reg_proba = ResidualDouble(reg_mean, reg_resid)  # doctest: +SKIP
    >>>
    >>> reg_proba.fit(X, y)  # doctest: +SKIP
    ResidualDouble(...)
    >>> y_pred = reg_proba.predict_proba(X)  # doctest: +SKIP
    >>> crps_metric = CRPS()  # doctest: +SKIP
    >>> plot_crossplot_loss(y, y_pred, crps_metric)  # doctest: +SKIP
    """
    _check_soft_dependencies("matplotlib")

    from matplotlib import pyplot

    losses = metric.evaluate_by_index(y_true, y_pred)
    loss_vals = losses.values.flatten()
    total_loss = np.mean(loss_vals).round(2)
    total_loss_std = np.std(loss_vals) / np.sqrt(len(loss_vals))
    total_loss_std = total_loss_std.round(2)

    overall = f"{total_loss} +/- {total_loss_std} sterr of mean"

    if ax is None:
        _, ax = pyplot.subplots()

    ax.plot(y_true, losses, "y_")

    ax.set_title(f"mean {metric.name}: {overall}")

    ax.set_xlabel(r"Correct label $y_i$")
    ax.set_ylabel(metric.name + r"($y_i$, $\widehat{y}_i$)")

    ax.tick_params(colors="y")

    return ax



"""Reducer to supervised regression - fit on uncensored."""

__author__ = ["fkiraly"]

from skpro.regression.base._delegate import _DelegatedProbaRegressor


class FitUncensored(_DelegatedProbaRegressor):
    """Reduction to tabular probabilistic regression - fit on uncensored subsample.

    Simple baseline reduction strategy for predictive survival analysis,
    fits a probabilistic regressor on the uncensored subsample of the data.

    In ``fit``, drops all observations that are right censored,
    and fits a probabilistic regressor on the remaining data.

    In ``predict``, uses the ``predict`` method of the fitted regressor.

    Parameters
    ----------
    estimator : skpro regressor, BaseProbaRegressor descendant
        probabilistic regressor to fit on uncensored subsample

    Attributes
    ----------
    estimator_ : skpro regressor, BaseProbaRegressor descendant
        fitted probabilistic regressor, fit on uncensored subsample
        clone of estimator
    """

    _tags = {"capability:survival": True}

    _delegate_name = "estimator_"

    def __init__(self, estimator):
        self.estimator = estimator

        super().__init__()

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Changes state to "fitted" = sets is_fitted flag to True

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.
        """
        if C is None:
            self.estimator_ = self.estimator.clone()
            self.estimator_.fit(X, y)
            return self

        uncensored_index = C.index[C.iloc[:, 0] == 0].tolist()

        X_uncensored = X.loc[uncensored_index, :]
        y_uncensored = y.loc[uncensored_index, :]

        self.estimator_ = self.estimator.clone()
        self.estimator_.fit(X_uncensored, y_uncensored)

        return self

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skpro.regression.bootstrap import BootstrapRegressor
        from skpro.regression.residual import ResidualDouble

        param1 = {"estimator": ResidualDouble.create_test_instance()}
        param2 = {"estimator": BootstrapRegressor.create_test_instance()}

        return [param1, param2]



"""Utility functions for adapting to sklearn."""

import numpy as np


def prep_skl_df(df, copy_df=False):
    """Make df compatible with sklearn input expectations.

    Changes:
    turns column index into a list of strings

    Parameters
    ----------
    df : pd.DataFrame
        list of indices to sample from
    copy_df : bool, default=False
        whether to mutate df or return a copy
        if False, index of df is mutated
        if True, original df is not mutated. If index is not a list of strings,
        a copy is made and the copy is mutated. Otherwise, the original df is returned.
    """
    cols = df.columns
    str_cols = cols.astype(str)

    if not np.all(str_cols == cols):
        if copy_df:
            df = df.copy()
        df.columns = str_cols

    return df



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""I.i.d. sample from a distribution."""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd

from skpro.distributions.base import BaseDistribution


class IID(BaseDistribution):
    r"""An i.i.d. sample of a given distribution.

    Constructed with a scalar or row distribution.

    Broadcasts to an i.i.d. sample based on index and columns.

    If ``distribution`` is scalar, broadcasts across index and columns.

    If ``distribution`` is a row distribution, ``columns`` must be
    same number of rows if passed; broadcasts across index.

    Parameters
    ----------
    distribution : skpro distribution - scalar, or shape (1, d)
        mean of the normal distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> import pandas as pd
    >>> from skpro.distributions.compose import IID
    >>> from skpro.distributions.normal import Normal

    >>> n = Normal(mu=0, sigma=1)
    >>> index = pd.Index([1, 2, 4, 6])
    >>> n_iid = IID(n, index=index)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["mean", "var", "energy", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "composite",
    }

    def __init__(self, distribution, index=None, columns=None):
        self.distribution = distribution

        dist_scalar = distribution.ndim == 0
        dist_cols = distribution.columns

        # set _bc_cols - do we broadcast rows?
        if dist_scalar:
            self._bc_cols = True
        else:  # not dist_scalar
            if distribution.shape[1] == 1 and columns is not None:
                self._bc_cols = True
            else:
                self._bc_cols = False

        # what is the index of self?
        if not dist_scalar and index is None:
            index = distribution.index
            assert len(index) == 1
        if dist_scalar and index is None:
            index = pd.RangeIndex(1)
        # else index is what was passed

        # what are columns of self?
        if dist_scalar and columns is None:
            columns = pd.RangeIndex(1)
        if not dist_scalar:
            if columns is None:
                columns = dist_cols
            if not len(dist_cols) == 1 and columns is not None:
                assert len(dist_cols) == len(columns)

        super().__init__(index=index, columns=columns)

        tags_to_clone = [
            "distr:measuretype",
            "capabilities:exact",
            "capabilities:approx",
        ]
        self.clone_tags(distribution, tags_to_clone)

    # TODO - use outer product once implemented, see issue #341
    # switch the method out to _broadcast_iid_future
    def _broadcast_iid(self, method, **kwargs):
        if self.ndim == 0:
            return getattr(self.distribution, method)(**kwargs)

        def _to_numpy(v):
            if isinstance(v, pd.DataFrame):
                return v.values
            return v

        kwargs = {k: _to_numpy(v) for k, v in kwargs.items()}

        # methods that return a single column always
        METHODS_TO_1D = ["energy"]

        if len(kwargs) == 0:
            one_dist_res = getattr(self.distribution, method)()
            if isinstance(one_dist_res, pd.DataFrame):
                one_dist_res = one_dist_res.values
            res = np.broadcast_to(one_dist_res, self.shape)
        else:
            if self._bc_cols:
                res = np.zeros(self.shape)
                for i, j in np.ndindex(self.shape):
                    kwargs_ij = {k: v[i, j] for k, v in kwargs.items()}
                    res[i, j] = getattr(self.distribution, method)(**kwargs_ij)
            elif method not in METHODS_TO_1D:
                res = np.zeros(self.shape)
                for i in range(self.shape[0]):
                    kwargs_i = {k: v[i] for k, v in kwargs.items()}
                    res[i] = getattr(self.distribution, method)(**kwargs_i)
            else:
                res = np.zeros((self.shape[0], 1))
                for i in range(self.shape[0]):
                    kwargs_i = {k: v[i] for k, v in kwargs.items()}
                    res[i] = getattr(self.distribution, method)(**kwargs_i)

        if method in METHODS_TO_1D and res.shape[1] > 1:
            res = np.mean(res, axis=1)[:, np.newaxis]
        return res

    # this should work once outer product is implemented
    def _broadcast_iid_future(self, method, **kwargs):
        one_dist_res = getattr(self.distribution, method)(**kwargs)
        if self.ndim == 0:
            return one_dist_res

        # methods that return a single column always
        METHODS_TO_1D = ["energy"]

        if isinstance(one_dist_res, pd.DataFrame):
            one_dist = one_dist_res.values

        if method in METHODS_TO_1D:
            target_shape = (self.shape[0], 1)
        else:
            target_shape = self.shape
        bc_res = np.broadcast_to(one_dist, target_shape)
        return bc_res

    def _iloc(self, rowidx=None, colidx=None):
        if is_scalar_notnone(rowidx) and is_scalar_notnone(colidx):
            return self._iat(rowidx, colidx)
        if is_scalar_notnone(rowidx):
            rowidx = pd.Index([rowidx])
        if is_scalar_notnone(colidx):
            colidx = pd.Index([colidx])

        if rowidx is not None:
            new_index = self.index[rowidx]
        else:
            new_index = self.index

        if colidx is not None:
            new_columns = self.columns[colidx]
        else:
            new_columns = self.columns

        if not self._bc_cols:
            distr = self.distribution.iloc[:, colidx]
        else:
            distr = self.distribution

        return IID(distribution=distr, index=new_index, columns=new_columns)

    def _iat(self, rowidx=None, colidx=None):
        if rowidx is None or colidx is None:
            raise ValueError("iat method requires both row and column index")
        self_subset = self.iloc[[rowidx], [colidx]]
        return type(self)(distribution=self_subset.distribution.iat[0, 0])

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        return self._broadcast_iid("mean")

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        return self._broadcast_iid("var")

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        return self._broadcast_iid("pdf", x=x)

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        return self._broadcast_iid("log_pdf", x=x)

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        return self._broadcast_iid("cdf", x=x)

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        return self._broadcast_iid("ppf", p=p)

    def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        return self._broadcast_iid("energy")

    def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        return self._broadcast_iid("energy", x=x)

    def sample(self, n_samples=None):
        """Sample from the distribution.

        Parameters
        ----------
        n_samples : int, optional, default = None
            number of samples to draw from the distribution

        Returns
        -------
        pd.DataFrame
            samples from the distribution

            * if ``n_samples`` is ``None``:
            returns a sample that contains a single sample from ``self``,
            in ``pd.DataFrame`` mtype format convention, with ``index`` and ``columns``
            as ``self``
            * if n_samples is ``int``:
            returns a ``pd.DataFrame`` that contains ``n_samples`` i.i.d.
            samples from ``self``, in ``pd-multiindex`` mtype format convention,
            with same ``columns`` as ``self``, and row ``MultiIndex`` that is product
            of ``RangeIndex(n_samples)`` and ``self.index``
        """
        # if delegate has ppf,
        dist_has_impl = self.distribution._has_implementation_of
        if dist_has_impl("_ppf") or dist_has_impl("ppf"):
            return super().sample(n_samples=n_samples)

        # else we sample manually, this will be less efficient due to loops
        if n_samples is None:
            n = 1
        else:
            n = n_samples
        target_shape = (n * self.shape[0], self.shape[1])
        samples = np.zeros(target_shape)

        if self._bc_cols:
            for i, j in np.ndindex(target_shape):
                samples[i, j] = self.distribution.sample()
        else:
            for i in range(target_shape[0]):
                samples[i] = self.distribution.sample()

        if n_samples is None:
            res_index = self.index
        else:
            res_index = pd.MultiIndex.from_product([range(n), self.index])
        return pd.DataFrame(samples, index=res_index, columns=self.columns)

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        from skpro.distributions import Normal

        n_scalar = Normal(mu=0, sigma=1)
        # array case examples
        params1 = {
            "distribution": n_scalar,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }

        n_row = Normal(mu=[1, 2], sigma=1, columns=pd.Index(["c", "d"]))
        params2 = {
            "distribution": n_row,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),  # this should override n_row.columns
        }

        # scalar case example - corner case, iid does nothing
        params3 = {"distribution": n_scalar}
        return [params1, params2, params3]


def is_scalar_notnone(obj):
    """Check if obj is scalar and not None."""
    return obj is not None and np.isscalar(obj)



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Alpha probability distribution."""

__author__ = ["SaiReavanth25"]

import pandas as pd
from scipy.stats import alpha, rv_continuous

from skpro.distributions.adapters.scipy import _ScipyAdapter


class Alpha(_ScipyAdapter):
    r"""Alpha distribution.

    Most methods wrap ``scipy.stats.alpha``.

    The alpha distribution is characterized by its shape parameter :math:`\a`,
    which determines its skewness and tail behavior.
    It is often used for modeling data with heavy right tails,
    unlike the Gaussian distribution(which is symmetric and bell-shaped).

    The probability density function (PDF) of the Alpha distribution is given by:
    .. math::

        f(x) = \frac{1}{x^2 \Phi(a) \sqrt{2\pi}}
                \exp\left(-\frac{1}{2}\left(\frac{a - 1}{x}\right)^2\right)

    where:
       - :math:`a` is the shape parameter.
       - :math:`Phi` is the cumulative distribution function (CDF) of the
                    standard normal distribution.

    Parameters
    ----------
    a : float or array of float (1D or 2D), must be positive
        Shape parameter controlling skewness and tail behavior.
        Higher values result in heavier tails and greater skewness towards the right.
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions import Alpha

    >>> distr = Alpha(a=[[1, 2], [3, 4]])
    """

    _tags = {
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, a, index=None, columns=None):
        self.a = a

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return alpha

    def _get_scipy_param(self):
        a = self._bc_params["a"]

        return [a], {}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        params1 = {"a": [[2, 3], [4, 5]]}
        params2 = {
            "a": 3,
            "index": pd.Index([1, 2, 3]),
            "columns": pd.Index(["a", "b"]),
        }
        params3 = {"a": 2.5}

        return [params1, params2, params3]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Base classes for probability array distribution objects."""

__author__ = ["ShreeshaM07"]

__all__ = ["_BaseArrayDistribution"]

import numpy as np
import pandas as pd

from skpro.base import BaseObject
from skpro.distributions.base import BaseDistribution
from skpro.distributions.base._base import (
    _coerce_to_pd_index_or_none,
    is_scalar_notnone,
)


class _BaseArrayDistribution(BaseDistribution, BaseObject):
    """Base Array probability distribution."""

    def __init__(self, index=None, columns=None):
        self.index = _coerce_to_pd_index_or_none(index)
        self.columns = _coerce_to_pd_index_or_none(columns)

        super().__init__(index=index, columns=columns)

    def _loc(self, rowidx=None, colidx=None):
        if is_scalar_notnone(rowidx) and is_scalar_notnone(colidx):
            return self._at(rowidx, colidx)
        if is_scalar_notnone(rowidx):
            rowidx = pd.Index([rowidx])
        if is_scalar_notnone(colidx):
            colidx = pd.Index([colidx])

        if rowidx is not None:
            row_iloc = pd.Index(self.index.get_indexer_for(rowidx))
        else:
            row_iloc = None
        if colidx is not None:
            col_iloc = pd.Index(self.columns.get_indexer_for(colidx))
        else:
            col_iloc = None
        return self._iloc(rowidx=row_iloc, colidx=col_iloc)

    def _subset_params(self, rowidx, colidx, coerce_scalar=False):
        """Subset distribution parameters to given rows and columns.

        Parameters
        ----------
        rowidx : None, numpy index/slice coercible, or int
            Rows to subset to. If None, no subsetting is done.
        colidx : None, numpy index/slice coercible, or int
            Columns to subset to. If None, no subsetting is done.
        coerce_scalar : bool, optional, default=False
            If True, and the subsetted parameter is a scalar, coerce it to a scalar.

        Returns
        -------
        dict
            Dictionary with subsetted distribution parameters.
            Keys are parameter names of ``self``, values are the subsetted parameters.
        """
        params = self._get_dist_params()

        subset_param_dict = {}
        for param, val in params.items():
            if val is None:
                subset_param_dict[param] = None
                continue
            arr = val
            arr_shape = 2
            # when rowidx and colidx are integer while plotting
            if coerce_scalar:
                arr = arr[rowidx][colidx]
                subset_param_dict[param] = arr
                continue
            # subset the 2D distributions
            if arr_shape == 2 and rowidx is not None:
                _arr_shift = []
                if rowidx.values is not None and colidx is None:
                    rowidx_list = rowidx.values
                    for row in rowidx:
                        _arr_shift.append(arr[row])

                elif rowidx.values is not None and colidx.values is not None:
                    rowidx_list = rowidx.values
                    colidx_list = colidx.values
                    for row in rowidx_list:
                        _arr_shift_row = []
                        for col in colidx_list:
                            _arr_shift_row.append(arr[row][col])
                        _arr_shift.append(_arr_shift_row)
                arr = _arr_shift

            if arr_shape == 2 and rowidx is None:
                _arr_shift = []
                if colidx is not None:
                    colidx_list = colidx.values
                    for row in range(len(arr)):
                        _arr_shift_row = []
                        for col in colidx_list:
                            _arr_shift_row.append(arr[row][col])
                        _arr_shift.append(_arr_shift_row)
                    arr = _arr_shift

            subset_param_dict[param] = arr
        return subset_param_dict

    def _iloc(self, rowidx=None, colidx=None):
        if is_scalar_notnone(rowidx) and is_scalar_notnone(colidx):
            return self._iat(rowidx, colidx)
        if is_scalar_notnone(rowidx):
            rowidx = pd.Index([rowidx])
        if is_scalar_notnone(colidx):
            colidx = pd.Index([colidx])

        if rowidx is not None:
            rowidx = pd.Index(rowidx)
        if colidx is not None:
            colidx = pd.Index(colidx)

        subset_params = self._subset_params(rowidx=rowidx, colidx=colidx)

        def subset_not_none(idx, subs):
            if subs is not None:
                return idx.take(pd.Index(subs))
            else:
                return idx

        index_subset = subset_not_none(self.index, rowidx)
        columns_subset = subset_not_none(self.columns, colidx)

        sk_distr_type = type(self)
        return sk_distr_type(
            index=index_subset,
            columns=columns_subset,
            **subset_params,
        )

    def _check_single_arr_distr(self, value):
        return (
            isinstance(value[0], int)
            or isinstance(value[0], np.integer)
            or isinstance(value[0], float)
            or isinstance(value[0], np.floating)
        )

    def _get_bc_params_dict(
        self, dtype=None, oned_as="row", return_shape=False, **kwargs
    ):
        """Fully broadcast dict of parameters given param shapes and index, columns.

        Parameters
        ----------
        kwargs : float, int, array of floats, or array of ints (1D or 2D)
            Distribution parameters that are to be made broadcastable. If no positional
            arguments are provided, all parameters of `self` are used except for `index`
            and `columns`.
        dtype : str, optional
            broadcasted arrays are cast to all have datatype `dtype`. If None, then no
            datatype casting is done.
        oned_as : str, optional, "row" (default) or "col"
            If 'row', then 1D arrays are treated as row vectors. If 'column', then 1D
            arrays are treated as column vectors.
        return_shape : bool, optional, default=False
            If True, return shape tuple, and a boolean tuple
            indicating which parameters are scalar.

        Returns
        -------
        dict of float or integer arrays
            Each element of the tuple represents a different broadcastable distribution
            parameter.
        shape : Tuple, only returned if ``return_shape`` is True
            Shape of the broadcasted parameters.
            Pair of row/column if not scalar, empty tuple if scalar.
        is_scalar : Tuple of bools, only returned if ``return_is_scalar`` is True
            Each element of the tuple is True if the corresponding parameter is scalar.
        """
        number_of_params = len(kwargs)
        if number_of_params == 0:
            # Handle case where no positional arguments are provided
            kwargs = self._get_dist_params()
            number_of_params = len(kwargs)

        # def row_to_col(arr):
        #     """Convert 1D arrays to 2D col arrays, leave 2D arrays unchanged."""
        #     if arr.ndim == 1 and oned_as == "col":
        #         return arr.reshape(-1, 1)
        #     return arr

        # kwargs_as_np = {k: row_to_col(np.array(v)) for k, v in kwargs.items()}
        kwargs_as_np = {k: v for k, v in kwargs.items()}

        if hasattr(self, "index") and self.index is not None:
            kwargs_as_np["index"] = self.index.to_numpy().reshape(-1, 1)
        if hasattr(self, "columns") and self.columns is not None:
            kwargs_as_np["columns"] = self.columns.to_numpy()

        bc_params = self.get_tags()["broadcast_params"]

        if bc_params is None:
            bc_params = kwargs_as_np.keys()

        args_as_np = [kwargs_as_np[k] for k in bc_params]

        if all(self._check_single_arr_distr(value) for value in kwargs_as_np.values()):
            # Convert all values in kwargs_as_np to np.array
            kwargs_as_np = {key: np.array(value) for key, value in kwargs_as_np.items()}
            shape = ()

            if return_shape:
                is_scalar = tuple([True] * (len(args_as_np) - 2))
                # print(kwargs_as_np,shape,is_scalar)
                return kwargs_as_np, shape, is_scalar
            return kwargs_as_np

        shape = (len(args_as_np[0]), len(args_as_np[0][0]))
        # create broadcast_array which will be same shape as the original bins
        # without considering the inner np.array containing the values of the bin edges
        # and bin masses. This will later get replaced by the values after broadcasting
        # index and columns.
        broadcast_array = np.arange(len(args_as_np[0]) * len(args_as_np[0][0])).reshape(
            shape
        )

        index_column_broadcast = [broadcast_array] * (len(args_as_np) - 2)
        index_column_broadcast.append(kwargs_as_np["index"])
        index_column_broadcast.append(kwargs_as_np["columns"])

        bc = list(np.broadcast_arrays(*index_column_broadcast))
        if dtype is not None:
            bc = [array.astype(dtype) for array in bc]

        for i in range(len(bc) - 2):
            bc[i] = args_as_np[i]

        for i, k in enumerate(bc_params):
            kwargs_as_np[k] = bc[i]

        if return_shape:
            is_scalar = tuple([False] * (len(args_as_np) - 2))
            # print(kwargs_as_np,shape,is_scalar)
            return kwargs_as_np, shape, is_scalar
        return kwargs_as_np



"""Example generation for testing.

Exports examples of in-memory data containers, useful for testing as fixtures.

Examples come in clusters, tagged by scitype: str, index: int, and metadata: dict.

All examples with the same index are considered "content-wise the same", i.e.,
representing the same abstract data object. They differ by mtype, i.e.,
machine type, which is the specific in-memory representation.

If an example returns None, it indicates that representation
with that specific mtype is not possible.

If the tag "lossy" is True, the representation is considered incomplete,
e.g., metadata such as column names are missing.

Types of tests that can be performed with these examples:

* the mtype and scitype of the example should be correctly inferred by checkers.
* the metadata of hte example should be correctly inferred by checkers.
* conversions from non-lossy representations to any other ones
  should yield the element exactly, identically, for examples of the same index.
"""

import numpy as np
import pandas as pd

from skpro.datatypes._base import BaseExample

###
# example 0: univariate


class _ProbaUniv(BaseExample):
    _tags = {
        "scitype": "Proba",
        "index": 0,
        "metadata": {
            "is_univariate": True,
            "is_empty": False,
            "has_nans": False,
        },
    }


class _ProbaUnivPredQ(_ProbaUniv):
    _tags = {
        "mtype": "pred_quantiles",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        pred_q = pd.DataFrame({0.2: [1, 2, 3], 0.6: [2, 3, 4]})
        pred_q.columns = pd.MultiIndex.from_product([["foo"], [0.2, 0.6]])

        return pred_q


class _ProbaUnivPredInt(_ProbaUniv):
    _tags = {
        "mtype": "pred_interval",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        # we need to use this due to numerical inaccuracies
        # from the binary based representation
        pseudo_0_2 = 2 * np.abs(0.6 - 0.5)

        pred_int = pd.DataFrame({0.2: [1, 2, 3], 0.6: [2, 3, 4]})
        pred_int.columns = pd.MultiIndex.from_tuples(
            [("foo", 0.6, "lower"), ("foo", pseudo_0_2, "upper")]
        )

        return pred_int


###
# example 1: multi


class _ProbaMulti(BaseExample):
    _tags = {
        "scitype": "Proba",
        "index": 1,
        "metadata": {
            "is_univariate": False,
            "is_empty": False,
            "has_nans": False,
        },
    }


class _ProbaMultiPredQ(_ProbaMulti):
    _tags = {
        "mtype": "pred_quantiles",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        pred_q = pd.DataFrame(
            {0.2: [1, 2, 3], 0.6: [2, 3, 4], 42: [5, 3, -1], 46: [5, 3, -1]}
        )
        pred_q.columns = pd.MultiIndex.from_product([["foo", "bar"], [0.2, 0.6]])

        return pred_q


class _ProbaMultiPredInt(_ProbaMulti):
    _tags = {
        "mtype": "pred_interval",
        "python_dependencies": None,
        "lossy": False,
    }

    def build(self):
        # we need to use this due to numerical inaccuracies
        # from the binary based representation
        pseudo_0_2 = 2 * np.abs(0.6 - 0.5)

        pred_int = pd.DataFrame(
            {0.2: [1, 2, 3], 0.6: [2, 3, 4], 42: [5, 3, -1], 46: [5, 3, -1]}
        )
        pred_int.columns = pd.MultiIndex.from_tuples(
            [
                ("foo", 0.6, "lower"),
                ("foo", pseudo_0_2, "upper"),
                ("bar", 0.6, "lower"),
                ("bar", pseudo_0_2, "upper"),
            ]
        )

        return pred_int



"""Retrieval utility for test scenarios."""
# copied from sktime. Should be jointly refactored to scikit-base.

__author__ = ["fkiraly"]

__all__ = ["retrieve_scenarios"]


from inspect import isclass

from skpro.tests.scenarios.scenarios_regressor_proba import scenarios_regressor_proba

scenarios = dict()
scenarios["regressor_proba"] = scenarios_regressor_proba


def retrieve_scenarios(obj, filter_tags=None):
    """Retrieve test scenarios for obj, or by estimator scitype string.

    Exactly one of the arguments obj, estimator_type must be provided.

    Parameters
    ----------
    obj : class or object, or string, or list of str.
        Which kind of estimator/object to retrieve scenarios for.
        If object, must be a class or object inheriting from BaseObject.
        If string(s), must be in registry.BASE_CLASS_REGISTER (first col)
            for instance 'classifier', 'regressor', 'transformer', 'forecaster'
    filter_tags: dict of (str or list of str), default=None
        subsets the returned objects as follows:
            each key/value pair is statement in "and"/conjunction
                key is tag name to sub-set on
                value str or list of string are tag values
                condition is "key must be equal to value, or in set(value)"

    Returns
    -------
    scenarios : list of objects, instances of BaseScenario
    """
    # if class, get scitypes from inference; otherwise, str or list of str
    if not isinstance(obj, str):
        if isclass(obj):
            if hasattr(obj, "get_class_tag"):
                estimator_type = obj.get_class_tag("object_type", "object")
            else:
                estimator_type = "object"
        else:
            if hasattr(obj, "get_tag"):
                estimator_type = obj.get_tag("object_type", "object", False)
            else:
                estimator_type = "object"
    else:
        estimator_type = obj

    # coerce to list, ensure estimator_type is list of str
    if not isinstance(estimator_type, list):
        estimator_type = [estimator_type]

    # now loop through types and retrieve scenarios
    scenarios_for_type = []
    for est_type in estimator_type:
        scens = scenarios.get(est_type)
        if scens is not None:
            scenarios_for_type += scenarios.get(est_type)

    # instantiate all scenarios by calling constructor
    scenarios_for_type = [x() for x in scenarios_for_type]

    # if obj was an object, filter to applicable scenarios
    if not isinstance(obj, str) and not isinstance(obj, list):
        scenarios_for_type = [x for x in scenarios_for_type if x.is_applicable(obj)]

    if filter_tags is not None:
        scenarios_for_type = [
            scen for scen in scenarios_for_type if _check_tag_cond(scen, filter_tags)
        ]

    return scenarios_for_type


def _check_tag_cond(obj, filter_tags=None):
    """Check whether object satisfies filter_tags condition.

    Parameters
    ----------
    obj: object inheriting from sktime BaseObject
    filter_tags: dict of (str or list of str), default=None
        subsets the returned objects as follows:
            each key/value pair is statement in "and"/conjunction
                key is tag name to sub-set on
                value str or list of string are tag values
                condition is "key must be equal to value, or in set(value)"

    Returns
    -------
    cond_sat: bool, whether estimator satisfies condition in filter_tags
    """
    if not isinstance(filter_tags, dict):
        raise TypeError("filter_tags must be a dict")

    cond_sat = True

    for key, value in filter_tags.items():
        if not isinstance(value, list):
            value = [value]
        cond_sat = cond_sat and obj.get_class_tag(key) in set(value)

    return cond_sat



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Truncated Normal probability distribution."""

__author__ = ["ShreeshaM07"]

import pandas as pd
from scipy.stats import rv_continuous, truncnorm

from skpro.distributions.adapters.scipy import _ScipyAdapter


class TruncatedNormal(_ScipyAdapter):
    """A truncated normal probability distribution.

    Most methods wrap ``scipy.stats.truncnorm``.
    It truncates the normal distribution at
    the abscissa ``l_trunc`` and ``r_trunc``.

    Note: The truncation parameters passed
    is internally shifted to be centred at
    mean and scaled by sigma.

    Parameters
    ----------
    mu : float or array of float (1D or 2D)
        mean of the normal distribution
    sigma : float or array of float (1D or 2D), must be positive
        standard deviation of the normal distribution
    l_trunc : float or array of float (1D or 2D)
        Left truncation abscissa.
    r_trunc : float or array of float (1D or 2D)
        Right truncation abscissa.
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.truncated_normal import TruncatedNormal

    >>> d = TruncatedNormal(\
            mu=[[0, 1], [2, 3], [4, 5]],\
            sigma= 1,\
            l_trunc= [[-0.1,0.5],[1.5,2.4],[4.1,5]],\
            r_trunc= [[0.8,2],[4,5],[5,7]]\
        )
    """

    _tags = {
        "capabilities:approx": ["energy", "pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, mu, sigma, l_trunc, r_trunc, index=None, columns=None):
        self.mu = mu
        self.sigma = sigma
        self.l_trunc = l_trunc
        self.r_trunc = r_trunc

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return truncnorm

    def _get_scipy_param(self):
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]
        l_trunc = self._bc_params["l_trunc"]
        r_trunc = self._bc_params["r_trunc"]

        # shift it to be centred at mu and sigma
        a = (l_trunc - mu) / sigma
        b = (r_trunc - mu) / sigma

        return [], {
            "loc": mu,
            "scale": sigma,
            "a": a,
            "b": b,
        }

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {
            "mu": [[0, 1], [2, 3], [4, 5]],
            "sigma": 1,
            "l_trunc": [[-0.1, 0.5], [1.5, 2.4], [4.1, 5]],
            "r_trunc": [[0.8, 2], [4, 5], [5, 7]],
        }
        params2 = {
            "mu": 0,
            "sigma": 1,
            "l_trunc": [-10, -5],
            "r_trunc": [5, 10],
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"mu": 1, "sigma": 2, "l_trunc": -3, "r_trunc": 5}
        return [params1, params2, params3]



"""Meta-strategy for online learning: refit on full data."""

__author__ = ["fkiraly"]
__all__ = ["OnlineRefitEveryN"]

import pandas as pd

from skpro.regression.base import _DelegatedProbaRegressor


class OnlineRefitEveryN(_DelegatedProbaRegressor):
    """Online regression strategy, updates only after N new data points are seen.

    In ``fit``, behaves like the wrapped regressor.

    In ``update``, runs the wrapped regressor's ``update`` only if
    at least N new data points have been seen since the last
    ``fit`` or ``update``.

    Can be combined with ``OnlineRefit`` to refit on the entire data,
    but only after N new data points have been seen.

    Parameters
    ----------
    estimator : skpro regressor, descendant of BaseProbaRegressor
        regressor to be update-refitted on all data, blueprint
    N : int, default=1
        number of new data points to see before updating the regressor

    Attributes
    ----------
    estimator_ : skpro regressor, descendant of BaseProbaRegressor
        clone of the regressor passed in the constructor, fitted on all data
    """

    _tags = {"capability:update": True}

    def __init__(self, estimator, N=1):
        self.estimator = estimator
        self.N = N

        super().__init__()

        tags_to_clone = [
            "capability:missing",
            "capability:survival",
        ]
        self.clone_tags(estimator, tags_to_clone)

        self.estimator_ = estimator.clone()

        self._X = None
        self._y = None
        self._C = None
        self.n_seen_since_last_update_ = 0

    def _update(self, X, y, C=None):
        """Update regressor with new batch of training data.

        State required:
            Requires state to be "fitted".

        Writes to self:
            Updates fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.

        Returns
        -------
        self : reference to self
        """
        X_pool = self._update_data(self._X, X)
        y_pool = self._update_data(self._y, y)
        C_pool = self._update_data(self._C, C)

        n_seen_now = len(X)
        n_seen_since_last_update = self.n_seen_since_last_update_ + n_seen_now

        if n_seen_since_last_update >= self.N:
            self.estimator_.update(X=X_pool, y=y_pool, C=C_pool)
            self._X = None
            self._y = None
            self._C = None
            self.n_seen_since_last_update_ = 0
        else:
            self.n_seen_since_last_update_ = n_seen_since_last_update

        return self

    def _update_data(self, X, X_new):
        """Update data with new batch of training data.

        Treats X_new as data with new indices, even if some indices overlap with X.

        Parameters
        ----------
        X : pandas DataFrame
        X_new : pandas DataFrame

        Returns
        -------
        X_updated : pandas DataFrame
            concatenated data, with reset index
        """
        if X is None and X_new is None:
            return None
        if X is None and X_new is not None:
            return X_new.reset_index(drop=True)
        if X is not None and X_new is None:
            return X.reset_index(drop=True)
        # else, both X and X_new are not None
        X_updated = pd.concat([X, X_new], ignore_index=True)
        return X_updated

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skbase.utils.dependencies import _check_estimator_deps
        from sklearn.linear_model import LinearRegression, Ridge

        from skpro.regression.residual import ResidualDouble
        from skpro.survival.coxph import CoxPH

        regressor = ResidualDouble(LinearRegression())

        params = [{"estimator": regressor}]

        if _check_estimator_deps(CoxPH, severity="none"):
            coxph = CoxPH()
            params.append({"estimator": coxph})
        else:
            ridge = Ridge()
            params.append({"estimator": ResidualDouble(ridge)})

        return params



"""Test scenarios for classification and regression.

Contains TestScenario concrete children to run in tests for classifiers/regressirs.
"""

__author__ = ["fkiraly"]

__all__ = ["scenarios_regressor_proba"]

from inspect import isclass

import numpy as np
import pandas as pd

from skpro.base import BaseObject
from skpro.tests.scenarios.scenarios import TestScenario


class _ProbaRegressorTestScenario(TestScenario, BaseObject):
    """Generic test scenario for classifiers."""

    def get_args(self, key, obj=None, deepcopy_args=True):
        """Return args for key. Can be overridden for dynamic arg generation.

        If overridden, must not have any side effects on self.args
            e.g., avoid assignments args[key] = x without deepcopying self.args first

        Parameters
        ----------
        key : str, argument key to construct/retrieve args for
        obj : obj, optional, default=None. Object to construct args for.
        deepcopy_args : bool, optional, default=True. Whether to deepcopy return.

        Returns
        -------
        args : argument dict to be used for a method, keyed by `key`
            names for keys need not equal names of methods these are used in
                but scripted method will look at key with same name as default
        """
        PREDICT_LIKE_FUNCTIONS = ["predict", "predict_var", "predict_proba"]
        # use same args for predict-like functions as for predict
        if key in PREDICT_LIKE_FUNCTIONS:
            key = "predict"

        return super().get_args(key=key, obj=obj, deepcopy_args=deepcopy_args)

    def is_applicable(self, obj):
        """Check whether scenario is applicable to obj.

        Parameters
        ----------
        obj : class or object to check against scenario

        Returns
        -------
        applicable: bool
            True if self is applicable to obj, False if not
        """

        def get_tag(obj, tag_name):
            if isclass(obj):
                return obj.get_class_tag(tag_name)
            else:
                return obj.get_tag(tag_name)

        # applicable only if object is a BaseProbaRegressor
        if not get_tag(obj, "object_type") == "regressor_proba":
            return False

        # if X has missing data, applicable only if can handle missing data
        has_missing_data = self.get_tag("X_missing", False, False)
        if has_missing_data and not get_tag(obj, "capability:missing"):
            return False

        return True


# as functions to ensure we can move to fixture-like structure later
def _load_diabetes_small():
    """Load diabetes dataset, return first 50 samples."""
    import pandas as pd
    from sklearn.datasets import load_diabetes

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    X = X.iloc[:50]
    y = y.iloc[:50]
    y = pd.DataFrame(y)

    return X, y


def _get_Xy_traintest():
    """Get X, y train/test data for regression tests.

    X, y are pandas.DataFrame, y is univariate.
    X has homogeneous dtype, and no missing data.

    Returns
    -------
    X_train : pandas.DataFrame, training data
    X_test : pandas.DataFrame, test data
    y_train : pandas.DataFrame, training labels
    y_test : pandas.DataFrame, test labels
    """
    from sklearn.model_selection import train_test_split

    X, y = _load_diabetes_small()
    X_train, X_test, y_train, y_test = train_test_split(X, y)

    return X_train, X_test, y_train, y_test


class ProbaRegressorBasic(_ProbaRegressorTestScenario):
    """Fit/predict with multivariate pandas mtype (same dtype), and labels y."""

    _tags = {
        "X_univariate": False,
        "X_missing": False,
        "y_univariate": True,
        "is_enabled": True,
    }

    @property
    def args(self):
        X_train, X_test, y_train, _ = _get_Xy_traintest()
        return {
            "fit": {"X": X_train, "y": y_train},
            "predict": {"X": X_test},
        }

    default_method_sequence = ["fit", "predict", "predict_proba"]
    default_arg_sequence = ["fit", "predict", "predict"]


class ProbaRegressorSurvival(_ProbaRegressorTestScenario):
    """Fit/predict with survival data, including a censoring dataset."""

    _tags = {
        "X_univariate": False,
        "X_missing": False,
        "y_univariate": True,
        "is_enabled": True,
    }

    @property
    def args(self):
        X_train, X_test, y_train, _ = _get_Xy_traintest()
        C_train = pd.DataFrame(
            np.random.choice([0, 1], size=len(y_train)),
            index=y_train.index,
            columns=y_train.columns,
        )
        return {
            "fit": {"X": X_train, "y": y_train, "C": C_train},
            "predict": {"X": X_test},
        }

    default_method_sequence = ["fit", "predict", "predict_proba"]
    default_arg_sequence = ["fit", "predict", "predict"]


# as a function to ensure we can move to fixture-like structure later
def _get_Xy_traintest_X_mixix_ynp():
    """Get X, y train/test data for regression tests.

    X is pandas.DataFrame, y is univariate numpy array.
    X has homogeneous dtype, and no missing data.

    X has mixed column index type (int and str).

    Returns
    -------
    X_train : pandas.DataFrame, training data
    X_test : pandas.DataFrame, test data
    y_train : pandas.DataFrame, training labels
    y_test : pandas.DataFrame, test labels
    """
    from sklearn.model_selection import train_test_split

    X, y = _load_diabetes_small()

    X.columns = ["age", "sex", "bmi", "bp", 1, 2, 3, 4, 5, 6]
    y = y.values

    X_train, X_test, y_train, y_test = train_test_split(X, y)

    return X_train, X_test, y_train, y_test


class ProbaRegressorXcolMixIxYnp(_ProbaRegressorTestScenario):
    """Fit/predict with multivariate pandas mtype, mixed col idx type."""

    _tags = {
        "X_univariate": False,
        "X_missing": False,
        "y_univariate": True,
        "is_enabled": True,
    }

    @property
    def args(self):
        X_train_mc, X_test_mc, y_train_mc, _ = _get_Xy_traintest_X_mixix_ynp()
        return {
            "fit": {"X": X_train_mc, "y": y_train_mc},
            "predict": {"X": X_test_mc},
        }

    default_method_sequence = ["fit", "predict", "predict_proba"]
    default_arg_sequence = ["fit", "predict", "predict"]


scenarios_regressor_proba = [
    ProbaRegressorBasic,
    ProbaRegressorXcolMixIxYnp,
    ProbaRegressorSurvival,
]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Beta probability distribution."""

__author__ = ["malikrafsan"]

import pandas as pd
from scipy.stats import beta, rv_continuous

from skpro.distributions.adapters.scipy import _ScipyAdapter


class Beta(_ScipyAdapter):
    r"""Beta distribution.

    Most methods wrap ``scipy.stats.beta``.

    The Beta distribution is parametrized by two shape parameters :math:`\alpha`
    and :math:`\beta`, such that the probability density function (PDF) is given by:

    .. math:: f(x) = \frac{x^{\alpha-1}(1-x)^{\beta-1}}{B(\alpha, \beta)}

    where :math:`B(\alpha, \beta)` is the beta function. The beta function
    is a normalization constant to ensure that the total probability is 1.

    Parameters
    ----------
    alpha : float or array of float (1D or 2D), must be positive
    beta : float or array of float (1D or 2D), must be positive
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.beta import Beta

    >>> d = Beta(beta=[[1, 1], [2, 3], [4, 5]], alpha=2)
    """

    _tags = {
        "capabilities:approx": ["energy", "pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, alpha, beta, index=None, columns=None):
        self.alpha = alpha
        self.beta = beta

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return beta

    def _get_scipy_param(self):
        alpha = self._bc_params["alpha"]
        beta = self._bc_params["beta"]

        return [], {"a": alpha, "b": beta}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"alpha": [[1, 1], [2, 3], [4, 5]], "beta": 3}
        params2 = {
            "alpha": 2,
            "beta": 3,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"alpha": 1.5, "beta": 2.1}

        return [params1, params2, params3]



"""Tests for quantile-parameterized distributions."""

import numpy as np
import pytest

from skpro.distributions.qpd import QPD_B, QPD_S, QPD_U
from skpro.tests.test_switch import run_test_for_class


@pytest.mark.skipif(
    not run_test_for_class(QPD_B),
    reason="run test only if softdeps are present and incrementally (if requested)",  #
)
def test_qpd_b_simple_use():
    """Test simple use of qpd with bounded mode."""
    qpd = QPD_B(
        alpha=0.2,
        qv_low=[1, 2],
        qv_median=[3, 4],
        qv_high=[5, 6],
        lower=0,
        upper=10,
    )

    qpd.mean()


@pytest.mark.skipif(
    not run_test_for_class(QPD_B),
    reason="run test only if softdeps are present and incrementally (if requested)",  #
)
def test_qpd_b_pdf():
    """Test pdf of qpd with bounded mode."""
    # these parameters should produce a uniform on -0.5, 0.5
    qpd_linear = QPD_B(
        alpha=0.2,
        qv_low=-0.3,
        qv_median=0,
        qv_high=0.3,
        lower=-0.5,
        upper=0.5,
    )
    x = np.linspace(-0.45, 0.45, 100)
    pdf_vals = [qpd_linear.pdf(x_) for x_ in x]
    np.testing.assert_allclose(pdf_vals, 1.0, rtol=1e-5)


@pytest.mark.skipif(
    not run_test_for_class(QPD_S),
    reason="run test only if softdeps are present and incrementally (if requested)",
)
def test_qpd_s_simple_use():
    """Test simple use of qpd with semi-bounded mode."""
    qpd = QPD_S(
        alpha=0.2,
        qv_low=[1, 2],
        qv_median=[3, 4],
        qv_high=[5, 6],
        lower=0,
    )

    qpd.mean()


@pytest.mark.skipif(
    not run_test_for_class(QPD_U),
    reason="run test only if softdeps are present and incrementally (if requested)",
)
def test_qpd_u_simple_use():
    """Test simple use of qpd with un-bounded mode."""
    qpd = QPD_U(
        alpha=0.2,
        qv_low=[1, 2],
        qv_median=[3, 4],
        qv_high=[5, 6],
    )

    qpd.mean()



"""Reducer to supervised regression - condition on uncensored."""

__author__ = ["fkiraly"]

import pandas as pd

from skpro.regression.base import BaseProbaRegressor


class ConditionUncensored(BaseProbaRegressor):
    """Reduction to tabular probabilistic regression - conditioning on uncensored.

    Simple baseline reduction strategy for predictive survival analysis.

    Fits a probabilistic regressor on X padded with censoring information C,
    in predict applies the fitted regressor to X padded with 0 (non-censored).

    In ``fit``, passes column concat of ``X`` and ``C`` to ``regressor.fit``.

    In ``predict_[method]``, calls ``regressor.predict_[method]``
    on ``X`` with an additional ``C``-like column, padded with 0,
    and returns the result.

    Parameters
    ----------
    estimator : skpro regressor, BaseProbaRegressor descendant
        probabilistic regressor to predict survival time from features

    Attributes
    ----------
    estimator_ : skpro regressor, BaseProbaRegressor descendant
        fitted probabilistic regressor, clone of ``regressor``
    """

    _tags = {"capability:survival": True}

    def __init__(self, estimator):
        self.estimator = estimator

        super().__init__()

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Changes state to "fitted" = sets is_fitted flag to True

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.
        """
        self._y_cols = y.columns

        X_and_C = self._get_padded_X(X, C)
        self.estimator_ = self.estimator.clone().fit(X_and_C, y)

        return self

    def _get_padded_X(self, X, C=None):
        """Get one-padded X to use in fit and predict methods."""
        X = X.copy()
        columns = self._y_cols
        index = X.index

        if C is None:
            C = pd.DataFrame(0, index=index, columns=columns)
        else:
            C = C.copy().astype("float")
        X_and_C = pd.concat([X, C], axis=1)
        return X_and_C

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        X_and_C = self._get_padded_X(X)
        y_pred = self.estimator_.predict(X_and_C)
        return y_pred

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y_pred : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        X_and_C = self._get_padded_X(X)
        y_pred = self.estimator_.predict_proba(X_and_C)
        return y_pred

    def _predict_interval(self, X, coverage):
        """Compute/return interval predictions.

        private _predict_interval containing the core logic,
            called from predict_interval and default _predict_quantiles

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : guaranteed list of float of unique values
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        X_and_C = self._get_padded_X(X)
        y_pred = self.estimator_.predict_interval(X_and_C, coverage=coverage)
        return y_pred

    def _predict_quantiles(self, X, alpha):
        """Compute/return quantile predictions.

        private _predict_quantiles containing the core logic,
            called from predict_quantiles and default _predict_interval

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : guaranteed list of float
            A list of probabilities at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
                second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
                at quantile probability in second col index, for the row index.
        """
        X_and_C = self._get_padded_X(X)
        y_pred = self.estimator_.predict_quantiles(X_and_C, alpha=alpha)
        return y_pred

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skpro.regression.bootstrap import BootstrapRegressor
        from skpro.regression.residual import ResidualDouble

        param1 = {"estimator": ResidualDouble.create_test_instance()}
        param2 = {"estimator": BootstrapRegressor.create_test_instance()}

        return [param1, param2]



"""Testing machine type converters for scitypes - convert_to utility."""

__author__ = ["fkiraly"]

import pytest

from skpro.datatypes._convert import convert_to
from skpro.datatypes._examples import get_examples
from skpro.tests.test_switch import run_test_module_changed
from skpro.utils import deep_equals

# hard-coded scitypes/mtypes to use in test_convert_to
#   easy to change in case the strings change
SCITYPES = ["Proba", "Table"]
MTYPES_PROBA = ["pred_interval", "pred_quantiles"]
MTYPES_TABLE = ["list_of_dict", "pd_Series_Table", "numpy2D"]


@pytest.mark.skipif(
    not run_test_module_changed(["skpro.datatypes", "skpro.utils"]),
    reason="Test only if skpro.datatypes has been changed",
)
def test_convert_to_simple():
    """Testing convert_to basic call works."""
    scitype = SCITYPES[0]

    from_fixt = get_examples(mtype=MTYPES_PROBA[1], as_scitype=scitype).get(0)
    # expectation is that the conversion is to mtype MTYPES_PROBA[0]
    exp_fixt = get_examples(mtype=MTYPES_PROBA[0], as_scitype=scitype).get(0)

    # carry out the conversion using convert_to
    converted = convert_to(from_fixt, to_type=MTYPES_PROBA[0], as_scitype=scitype)

    # compare expected output with actual output of convert_to
    msg = "convert_to basic call does not seem to work."
    assert deep_equals(converted, exp_fixt), msg


@pytest.mark.skipif(
    not run_test_module_changed(["skpro.datatypes", "skpro.utils"]),
    reason="Test only if skpro.datatypes has been changed",
)
def test_convert_to_without_scitype():
    """Testing convert_to call without scitype specification."""
    scitype = SCITYPES[0]

    from_fixt = get_examples(mtype=MTYPES_PROBA[1], as_scitype=scitype).get(0)
    # convert_to should recognize the correct scitype, otherwise same as above
    exp_fixt = get_examples(mtype=MTYPES_PROBA[0], as_scitype=scitype).get(0)

    # carry out the conversion using convert_to
    converted = convert_to(from_fixt, to_type=MTYPES_PROBA[0])

    # compare expected output with actual output of convert_to
    msg = "convert_to call without scitype does not seem to work."
    assert deep_equals(converted, exp_fixt), msg


@pytest.mark.skipif(
    not run_test_module_changed(["skpro.datatypes", "skpro.utils"]),
    reason="Test only if skpro.datatypes has been changed",
)
def test_convert_to_mtype_list():
    """Testing convert_to call to_type being a list, of same scitype."""
    # convert_to list
    target_list = MTYPES_TABLE[:2]
    scitype = SCITYPES[0]

    # example that is on the list
    from_fixt_on = get_examples(mtype=MTYPES_TABLE[1], as_scitype=scitype).get(0)
    # example that is not on the list
    from_fixt_off = get_examples(mtype=MTYPES_TABLE[2], as_scitype=scitype).get(0)

    # if on the list, result should be equal to input
    exp_fixt_on = get_examples(mtype=MTYPES_TABLE[1], as_scitype=scitype).get(0)
    # if off the list, result should be converted to mtype that is first on the list
    exp_fixt_off = get_examples(mtype=MTYPES_TABLE[0], as_scitype=scitype).get(0)

    # carry out the conversion using convert_to
    converted_on = convert_to(from_fixt_on, to_type=target_list)
    converted_off = convert_to(from_fixt_off, to_type=target_list)

    # compare expected output with actual output of convert_to
    msg = "convert_to call does not work with list for to_type."
    assert deep_equals(converted_on, exp_fixt_on), msg
    assert deep_equals(converted_off, exp_fixt_off), msg



"""Interface adapters to scikit-survival Cox-net model."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.survival.adapters.sksurv import _SksurvAdapter
from skpro.survival.base import BaseSurvReg


class CoxNet(_SksurvAdapter, BaseSurvReg):
    """Cox proportional hazards model with elastic net penalty.

    Direct interface to ``sksurv.linear_model.CoxnetSurvivalAnalysis``, by ``sebp``.

    Parameters
    ----------
    n_alphas : int, optional, default: 100
        Number of alphas along the regularization path.

    alphas : array-like or None, optional
        List of alphas where to compute the models.
        If ``None`` alphas are set automatically.

    alpha_min_ratio : float or { "auto" }, optional, default: "auto"
        Determines minimum alpha of the regularization path
        if ``alphas`` is ``None``. The smallest value for alpha
        is computed as the fraction of the data derived maximum
        alpha (i.e. the smallest value for which all
        coefficients are zero).

        If set to "auto", the value will depend on the
        sample size relative to the number of features.
        If ``n_samples > n_features``, the default value is 0.0001
        If ``n_samples <= n_features``, 0.01 is the default value.

    l1_ratio : float, optional, default: 0.5
        The ElasticNet mixing parameter, with ``0 < l1_ratio <= 1``.
        For ``l1_ratio = 0`` the penalty is an L2 penalty.
        For ``l1_ratio = 1`` it is an L1 penalty.
        For ``0 < l1_ratio < 1``, the penalty is a combination of L1 and L2.

    penalty_factor : array-like or None, optional
        Separate penalty factors can be applied to each coefficient.
        This is a number that multiplies alpha to allow differential
        shrinkage.  Can be 0 for some variables, which implies no shrinkage,
        and that variable is always included in the model.
        Default is 1 for all variables.

        Note: the penalty factors are internally rescaled to sum to
        ``n_features``, and the alphas sequence will reflect this change.

    normalize : boolean, optional, default: False
        If True, the features X will be normalized before optimization by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        ``sklearn.preprocessing.StandardScaler`` before calling ``fit``
        on an estimator with ``normalize=False``.

    copy_X : boolean, optional, default: True
        If ``True``, X will be copied; else, it may be overwritten.

    tol : float, optional, default: 1e-7
        The tolerance for the optimization: optimization continues
        until all updates are smaller than ``tol``.

    max_iter : int, optional, default: 100000
        The maximum number of iterations.

    verbose : bool, optional, default: False
        Whether to print additional information during optimization.

    Attributes
    ----------
    alphas_ : ndarray, shape=(n_alphas,)
        The actual sequence of alpha values used.

    alpha_min_ratio_ : float
        The inferred value of alpha_min_ratio.

    penalty_factor_ : ndarray, shape=(n_features,)
        The actual penalty factors used.

    coef_ : ndarray, shape=(n_features, n_alphas)
        Matrix of coefficients.

    offset_ : ndarray, shape=(n_alphas,)
        Bias term to account for non-centered features.

    deviance_ratio_ : ndarray, shape=(n_alphas,)
        The fraction of (null) deviance explained.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    Examples
    --------
    >>> from skpro.survival.linear import CoxNet  # doctest: +SKIP
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>> from sklearn.model_selection import train_test_split  # doctest: +SKIP
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)  # doctest: +SKIP

    >>> reg_proba = CoxNet()  # doctest: +SKIP
    >>> reg_proba.fit(X_train, y_train)  # doctest: +SKIP
    >>> y_pred = reg_proba.predict_proba(X_test)  # doctest: +SKIP
    """

    _tags = {"authors": ["sebp", "fkiraly"]}  # sebp credit for interfaced estimator

    def __init__(
        self,
        n_alphas=100,
        alphas=None,
        alpha_min_ratio="auto",
        l1_ratio=0.5,
        penalty_factor=None,
        normalize=False,
        copy_X=True,
        tol=1e-7,
        max_iter=100000,
        verbose=False,
    ):
        self.n_alphas = n_alphas
        self.alphas = alphas
        self.alpha_min_ratio = alpha_min_ratio
        self.l1_ratio = l1_ratio
        self.penalty_factor = penalty_factor
        self.normalize = normalize
        self.copy_X = copy_X
        self.tol = tol
        self.max_iter = max_iter
        self.verbose = verbose

        super().__init__()

    def _get_sksurv_class(self):
        """Getter of the sksurv class to be used for the adapter."""
        from sksurv.linear_model import CoxnetSurvivalAnalysis as _CoxNet

        return _CoxNet

    def _get_sksurv_object(self):
        """Abstract method to initialize sksurv object.

        The default initializes result of _get_sksurv_class
        with self.get_params.
        """
        cls = self._get_sksurv_class()
        params = self.get_params()
        params["fit_baseline_model"] = True  # required for predict_survival_function
        # and therefore for _predict_proba implementation to be valid
        return cls(**params)

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "n_alphas": 99,
            "alpha_min_ratio": 0.001,
            "l1_ratio": 0.4,
            "normalize": True,
            "tol": 1e-6,
            "max_iter": 99999,
            "verbose": True,
        }

        return [params1, params2]



"""Tests for probabilistic quantile and interval metrics."""
import pandas as pd
import pytest
from sklearn.datasets import load_diabetes
from sklearn.model_selection import train_test_split

from skpro.metrics._classes import (
    ConstraintViolation,
    EmpiricalCoverage,
    IntervalWidth,
    PinballLoss,
)
from skpro.regression.residual import ResidualDouble

quantile_metrics = [
    PinballLoss,
]

interval_metrics = [
    EmpiricalCoverage,
    ConstraintViolation,
    IntervalWidth,
]

all_metrics = quantile_metrics + interval_metrics


X, y = load_diabetes(return_X_y=True, as_frame=True)
y = pd.DataFrame(y)
X_train, X_test, y_train, y_test = train_test_split(X, y)

# fit - just once for all predict output methods
reg = ResidualDouble.create_test_instance()
reg.fit(X_train, y_train)
"""
Cases we need to test
score average = TRUE/FALSE
multivariable = TRUE/FALSE
multiscores = TRUE/FALSE

Data types
Univariate and single score
Univariate and multi score
Multivariate and single score
Multivariate and multiscor

For each of the data types we need to test with score average = T/F \
    and multioutput with "raw_values" and "uniform_average"
"""
quantile_pred_uni_s = reg.predict_quantiles(X_test, alpha=[0.5])
interval_pred_uni_s = reg.predict_interval(X_test, coverage=0.9)
quantile_pred_uni_m = reg.predict_quantiles(X_test, alpha=[0.05, 0.5, 0.95])
interval_pred_uni_m = reg.predict_interval(X_test, coverage=[0.7, 0.8, 0.9, 0.99])

y_test_uni = y_test

# replace this by multivariate prediction once available
y_train2 = y_train.copy()
y_train2.columns = ["foo"]

reg.fit(X_train, y_train2)
quantile_pred_multi_s = reg.predict_quantiles(X_test, alpha=[0.5])
interval_pred_multi_s = reg.predict_interval(X_test, coverage=0.9)
quantile_pred_multi_m = reg.predict_quantiles(X_test, alpha=[0.05, 0.5, 0.95])
interval_pred_multi_m = reg.predict_interval(X_test, coverage=[0.7, 0.8, 0.9, 0.99])

quantile_pred_multi_s = pd.concat([quantile_pred_multi_s, quantile_pred_uni_s], axis=1)
interval_pred_multi_s = pd.concat([interval_pred_multi_s, interval_pred_uni_s], axis=1)
quantile_pred_multi_m = pd.concat([quantile_pred_multi_m, quantile_pred_uni_m], axis=1)
interval_pred_multi_m = pd.concat([interval_pred_multi_m, interval_pred_uni_m], axis=1)

y_test2 = y_test.copy()
y_test2.columns = ["foo"]
y_test_multi = pd.concat([y_test2, y_test], axis=1)

# replace this end
#
# todo: no example of multivariate prediction yet
# for now, we use dummy data above
# once that changes, replace the below lines appropriately
#
# quantile_pred_multi_s = f_multi.predict_quantiles(fh=fh_multi, alpha=[0.5])
# interval_pred_multi_s = f_multi.predict_interval(fh=fh_multi, coverage=0.9)
# quantile_pred_multi_m =
# f_multi.predict_quantiles(fh=fh_multi, alpha=[0.05, 0.5, 0.95])
# interval_pred_multi_m = f_multi.predict_interval(
#     fh=fh_multi, coverage=[0.7, 0.8, 0.9, 0.99]
# )

uni_data = [
    quantile_pred_uni_s,
    interval_pred_uni_s,
    quantile_pred_uni_m,
    interval_pred_uni_m,
]

multi_data = [
    quantile_pred_multi_s,
    interval_pred_multi_s,
    quantile_pred_multi_m,
    interval_pred_multi_m,
]

quantile_data = [
    quantile_pred_uni_s,
    quantile_pred_uni_m,
    quantile_pred_multi_s,
    quantile_pred_multi_m,
]

interval_data = [
    interval_pred_uni_s,
    interval_pred_uni_m,
    interval_pred_multi_s,
    interval_pred_multi_m,
]


@pytest.mark.parametrize(
    "y_true, y_pred",
    list(zip([y_test_uni] * 4, uni_data)) + list(zip([y_test_multi] * 4, multi_data)),
)
@pytest.mark.parametrize("metric", all_metrics)
@pytest.mark.parametrize("multioutput", ["uniform_average", "raw_values"])
@pytest.mark.parametrize("score_average", [True, False])
def test_output(metric, score_average, multioutput, y_true, y_pred):
    """Test output is correct class and shape for given data."""
    loss = metric.create_test_instance()
    loss.set_params(score_average=score_average, multioutput=multioutput)

    eval_loss = loss(y_true, y_pred)
    index_loss = loss.evaluate_by_index(y_true, y_pred)

    no_vars = len(y_pred.columns.get_level_values(0).unique())
    no_scores = len(y_pred.columns.get_level_values(1).unique())

    if (
        0.5 in y_pred.columns.get_level_values(1)
        and loss.get_tag("scitype:y_pred") == "pred_interval"
        and y_pred.columns.nlevels == 2
    ):
        no_scores = no_scores - 1
        no_scores = no_scores / 2  # one interval loss per two quantiles given
        if no_scores == 0:  # if only 0.5 quant, no output to interval loss
            no_vars = 0

    if score_average and multioutput == "uniform_average":
        assert isinstance(eval_loss, float)
        assert isinstance(index_loss, pd.Series)

        assert len(index_loss) == y_pred.shape[0]

    if not score_average and multioutput == "uniform_average":
        assert isinstance(eval_loss, pd.Series)
        assert isinstance(index_loss, pd.DataFrame)

        # get two quantiles from each interval so if not score averaging
        # get twice number of unique coverages
        if (
            loss.get_tag("scitype:y_pred") == "pred_quantiles"
            and y_pred.columns.nlevels == 3
        ):
            assert len(eval_loss) == 2 * no_scores
        else:
            assert len(eval_loss) == no_scores

    if not score_average and multioutput == "raw_values":
        assert isinstance(eval_loss, pd.Series)
        assert isinstance(index_loss, pd.DataFrame)

        true_len = no_vars * no_scores

        if (
            loss.get_tag("scitype:y_pred") == "pred_quantiles"
            and y_pred.columns.nlevels == 3
        ):
            assert len(eval_loss) == 2 * true_len
        else:
            assert len(eval_loss) == true_len

    if score_average and multioutput == "raw_values":
        assert isinstance(eval_loss, pd.Series)
        assert isinstance(index_loss, pd.DataFrame)

        assert len(eval_loss) == no_vars


@pytest.mark.parametrize("Metric", quantile_metrics)
@pytest.mark.parametrize(
    "y_pred, y_true", list(zip(quantile_data, [y_test_uni] * 2 + [y_test_multi] * 2))
)
def test_evaluate_alpha_positive(Metric, y_pred, y_true):
    """Tests output when required quantile is present."""
    # 0.5 in test quantile data don't raise error.
    Loss = Metric.create_test_instance().set_params(alpha=0.5, score_average=False)
    res = Loss(y_true=y_true, y_pred=y_pred)
    assert len(res) == 1

    if all(x in y_pred.columns.get_level_values(1) for x in [0.5, 0.95]):
        Loss = Metric.create_test_instance().set_params(
            alpha=[0.5, 0.95], score_average=False
        )
        res = Loss(y_true=y_true, y_pred=y_pred)
        assert len(res) == 2


@pytest.mark.parametrize("Metric", quantile_metrics)
@pytest.mark.parametrize(
    "y_pred, y_true", list(zip(quantile_data, [y_test_uni] * 2 + [y_test_multi] * 2))
)
def test_evaluate_alpha_negative(Metric, y_pred, y_true):
    """Tests whether correct error raised when required quantile not present."""
    with pytest.raises(ValueError):
        # 0.3 not in test quantile data so raise error.
        Loss = Metric.create_test_instance().set_params(alpha=0.3)
        res = Loss(y_true=y_true, y_pred=y_pred)  # noqa



"""Base classes for probabilistic regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["BaseProbaRegressor", "_DelegatedProbaRegressor"]

from skpro.regression.base._base import BaseProbaRegressor
from skpro.regression.base._delegate import _DelegatedProbaRegressor



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Weibull probability distribution."""

__author__ = ["malikrafsan"]

import numpy as np
import pandas as pd
from scipy.special import gamma

from skpro.distributions.base import BaseDistribution


class Weibull(BaseDistribution):
    r"""Weibull distribution.

    The Weibull distribution is parametrized by scale parameter :math:`\lambda`,
    and shape parameter :math:`k`, such that the cdf is given by:

    .. math:: F(x) = 1 - \exp\left(-\left(\frac{x}{\lambda}\right)^k\right)

    The scale parameter :math:`\lambda` is represented by the parameter ``scale``,
    and the shape parameter :math:`k` by the parameter ``k``.

    Parameters
    ----------
    scale : float or array of float (1D or 2D), must be positive
        scale parameter of the distribution
    k : float or array of float (1D or 2D), must be positive
        shape parameter of the distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.weibull import Weibull

    >>> w = Weibull(scale=[[1, 1], [2, 3], [4, 5]], k=1)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, scale, k, index=None, columns=None):
        self.scale = scale
        self.k = k

        super().__init__(index=index, columns=columns)

    def _mean(self):
        r"""Return expected value of the distribution.

        For Weibull distribution, expectation is given by,
        :math:`\lambda \Gamma (1+\frac{1}{k})`

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        scale = self._bc_params["scale"]
        k = self._bc_params["k"]
        mean_arr = scale * gamma(1 + 1 / k)
        return mean_arr

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        For Weibull distribution, variance is given by
        :math:`\lambda^2 \left( \Gamma(1+\frac{2}{k}) - \Gamma^2(1+\frac{1}{k}) \right)`

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        scale = self._bc_params["scale"]
        k = self._bc_params["k"]

        left_gamma = gamma(1 + 2 / k)
        right_gamma = gamma(1 + 1 / k) ** 2
        var_arr = scale**2 * (left_gamma - right_gamma)
        return var_arr

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        k = self._bc_params["k"]
        scale = self._bc_params["scale"]

        pdf_arr = (k / scale) * (x / scale) ** (k - 1) * np.exp(-((x / scale) ** k))
        pdf_arr = pdf_arr * (x >= 0)  # if x < 0, pdf = 0
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        k = self._bc_params["k"]
        scale = self._bc_params["scale"]

        lpdf_arr = np.log(k / scale) + (k - 1) * np.log(x / scale) - (x / scale) ** k
        lpdf_arr = lpdf_arr + np.log(x >= 0)  # if x < 0, pdf = 0, so log pdf = -inf
        return lpdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        k = self._bc_params["k"]
        scale = self._bc_params["scale"]

        cdf_arr = 1 - np.exp(-((x / scale) ** k))
        cdf_arr = cdf_arr * (x >= 0)  # if x < 0, cdf = 0
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        k = self._bc_params["k"]
        scale = self._bc_params["scale"]

        ppf_arr = scale * (-np.log(1 - p)) ** (1 / k)
        return ppf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"scale": [[1, 1], [2, 3], [4, 5]], "k": 1}
        params2 = {
            "scale": 1,
            "k": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"scale": 2, "k": 3}

        return [params1, params2, params3]
